
/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>

#include <debug/log.h>
#include <debug/memory.h>
#include <debug/hex.h>
#include <ber/ber.h>

#include <tinysnmp/compat/inttypes.h>

#include "config.h"
#include "mod.h"
#include "snmp.h"

#define ADDR(x) (x) & 0x000000ff,((x) & 0x0000ff00) >> 8,((x) & 0x00ff0000) >> 16,((x) & 0xff000000) >> 24

#define NB_OF(x) (sizeof (x) / sizeof ((x)[0]))

/* maximum UDP datagram size */
#define UDP_DATAGRAM_SIZE 65535

typedef struct
{
   const char *logfile;			/* location of log file             */
   int loglevel;				/* logging system verbosity         */
   int daemon;					/* are we a daemon?                 */
   const char *configfile;		/* location of configuration file   */
} options_t;

static volatile int finished = 0;
static volatile int rotated = 0;

static void catch_signals (int sig)
{
   if (sig == SIGUSR1)
	 rotated = 1;
   else
	 finished = 1;
}

/* send a packet to addr (if possible) */
static int send_packet (int fd,const ber_t *e,const struct sockaddr *addr)
{
   int flags = MSG_DONTWAIT;

#ifdef MSG_NOSIGNAL
   flags |= MSG_NOSIGNAL;
#endif	/* #ifdef MSG_NOSIGNAL */

#if defined(MSG_CONFIRM) && !defined(COMPAT22)
   flags |= MSG_CONFIRM;	/* this is good, but only Linux 2.3+ supports it (see sendto(2)) */
#endif	/* #if defined(MSG_CONFIRM) && !defined(COMPAT22) */

#ifdef DEBUG
   log_printf (LOG_DEBUG,"Sending packet (%" PRIu32 " bytes) to %d.%d.%d.%d:%d\n",
			   e->offset,
			   ADDR(((const struct sockaddr_in *) addr)->sin_addr.s_addr),
			   ((const struct sockaddr_in *) addr)->sin_port);
#endif	/* #ifdef DEBUG */

   return (sendto (fd,e->buf + e->size - e->offset,e->offset,flags,addr,sizeof (struct sockaddr)));
}
/* try to send an error message to addr */
static void send_error (int fd,ber_t *e,const snmp_pdu_t *pdu,const struct sockaddr *addr,uint8_t status,uint32_t index)
{
   if (!encode_get_response (e,pdu,status,index))
	 send_packet (fd,e,addr);
}

static int oscmp (octet_string_t *a,const char *b)
{
   size_t len = strlen (b);
   if (a->len < len)
	 return (-1);
   else if (a->len > len)
	 return (1);
   else return (memcmp (a->buf,b,len));
}

/* process an incoming packet */
static void process_incoming (int fd,ber_t *ber,const config_t *conf)
{
   int i;
   struct sockaddr addr;
   socklen_t len = sizeof (struct sockaddr);
   int result,flags = MSG_WAITALL;
   snmp_pdu_t pdu;
   uint32_t size = ber->size;

#ifdef MSG_NOSIGNAL
   flags |= MSG_NOSIGNAL;
#endif	/* #ifdef MSG_NOSIGNAL */

   /* if recvfrom() failed */
   if ((result = recvfrom (fd,ber->buf,ber->size + 1,flags,&addr,&len)) <= 0)
	 {
		log_printf (LOG_WARNING,
					"reject packet from %d.%d.%d.%d:%d: recvfrom() failed or returned 0\n",
					ADDR (((const struct sockaddr_in *) &addr)->sin_addr.s_addr),
					((const struct sockaddr_in *) &addr)->sin_port);
		return;
	 }

   /* make sure that the client is in the list of allowed hosts */
   for (i = 0; i < conf->n; i++)
	 if ((((const struct sockaddr_in *) &addr)->sin_addr.s_addr & conf->allow[i].netmask) == conf->allow[i].address)
	   break;

   if (i == conf->n)
	 {
		log_printf (LOG_WARNING,
					"reject packet from %d.%d.%d.%d:%d: not in list of allowed clients\n",
					ADDR (((const struct sockaddr_in *) &addr)->sin_addr.s_addr),
					((const struct sockaddr_in *) &addr)->sin_port);
		return;
	 }

   /* packet was truncated */
   if (result > ber->size)
	 {
		log_printf (LOG_WARNING,
					"reject packet from %d.%d.%d.%d:%d: incoming packet too big\n",
					ADDR (((const struct sockaddr_in *) &addr)->sin_addr.s_addr),
					((const struct sockaddr_in *) &addr)->sin_port);
		return;
	 }

   ber->size = result;

#ifdef DEBUG
   log_printf (LOG_DEBUG,
			   "received packet from %d.%d.%d.%d:%d\n",
			   ADDR (((const struct sockaddr_in *) &addr)->sin_addr.s_addr),
			   ((const struct sockaddr_in *) &addr)->sin_port);
   hexdump (LOG_DEBUG,ber->buf + ber->size - ber->offset,ber->offset);
#endif	/* #ifdef DEBUG */

   /* decode the packet */
   if ((result = decode_pdu (&pdu,ber)) < 0)
	 {
		log_printf (LOG_WARNING,
					"reject packet from %d.%d.%d.%d:%d: decoding of packet failed\n",
					ADDR (((const struct sockaddr_in *) &addr)->sin_addr.s_addr),
					((const struct sockaddr_in *) &addr)->sin_port);
		return;
	 }

   /* check community string */
   if (oscmp (&pdu.community,conf->community))
	 {
		log_printf (LOG_WARNING,
					"reject packet from %d.%d.%d.%d:%d: community string mismatch\n",
					ADDR (((const struct sockaddr_in *) &addr)->sin_addr.s_addr),
					((const struct sockaddr_in *) &addr)->sin_port);
error_after_decode:
		if (pdu.community.len) mem_free (pdu.community.buf);
		for (i = 0; i < pdu.n; i++) mem_free (pdu.oid[i]);
		mem_free (pdu.oid);
		return;
	 }

   ber->size = size;
   ber->offset = 0;

   /* try to encode the packet. if it fails, send a genErr(5) error message back */
   if (encode_get_response (ber,&pdu,noError,0) < 0)
	 {
		log_printf (LOG_WARNING,
					"reject packet from %d.%d.%d.%d:%d: unable to encode response\n",
					ADDR (((const struct sockaddr_in *) &addr)->sin_addr.s_addr),
					((const struct sockaddr_in *) &addr)->sin_port);
		send_error (fd,ber,&pdu,&addr,genErr,0);
		goto error_after_decode;
	 }

   /* try to send the packet */
   if ((result = send_packet (fd,ber,&addr)) <= 0)
	 {
		log_printf (LOG_WARNING,
					"reply to %s from %d.%d.%d.%d:%d failed\n",
					pdu.type == BER_GetRequest ? "get-request" : "get-next-request",
					ADDR (((const struct sockaddr_in *) &addr)->sin_addr.s_addr),
					((const struct sockaddr_in *) &addr)->sin_port);
		goto error_after_decode;
	 }

   log_printf (LOG_NORMAL,
			   "replied to %s from %d.%d.%d.%d:%d\n",
			   pdu.type == BER_GetRequest ? "get-request" : "get-next-request",
			   ADDR (((const struct sockaddr_in *) &addr)->sin_addr.s_addr),
			   ((const struct sockaddr_in *) &addr)->sin_port);

   if (pdu.community.len) mem_free (pdu.community.buf);
   for (i = 0; i < pdu.n; i++) mem_free (pdu.oid[i]);
   if (pdu.n) mem_free (pdu.oid);

#ifdef DEBUG
   log_printf (LOG_DEBUG,
			   "sent packet to %d.%d.%d.%d:%d:\n",
			   ADDR (((const struct sockaddr_in *) &addr)->sin_addr.s_addr),
			   ((const struct sockaddr_in *) &addr)->sin_port);
   hexdump (LOG_DEBUG,ber->buf + ber->size - result,result);
#endif	/* #ifdef DEBUG */
}

#define usage(p) help_stub(p,0)
#define help(p) help_stub(p,1)
static void help_stub (const char *progname,int verbose)
{
   fprintf (stderr,
			"usage: %s [OPTIONS] <configfile>\n"
			"       %s -h | --help\n",
			progname,progname);

   if (verbose)
	 {
		fprintf (stderr,
				 "\n"
				 "where OPTIONS can be any of the following:\n"
				 "\n"
				 "%s"
				 "\n"
				 "    -f | --logfile=FILENAME     Location of log file. By default, messages will be\n"
				 "                                logged to the console.\n"
				 "\n"
				 "    -l | --loglevel=LEVEL       The following levels can be specified (in increasing\n"
				 "                                order of verbosity):\n"
				 "\n"
				 "                                    quiet     No messages will be logged\n"
				 "                                    errors    Only errors will be logged\n"
				 "                                    warnings  Warnings & errors will be logged\n"
				 "                                    normal    Default messages and all of the above\n"
				 "                                    verbose   Verbose messages will also be logged\n"
				 "                                    debug     Log debug messages as well\n"
				 "                                    noisy     Noisy debug messages will be logged as well\n"
				 "\n"
				 "                                If a log file is specified, ORCA defaults to `verbose'',\n"
				 "                                otherwise, `noisy'' is selected.\n"
				 "\n"
				 "    -d | --daemon               Run in the background. You need to specify a log file if\n"
				 "                                you want to use this option.\n"
				 "\n"
				 "    -h | --help                 Show this help message.\n"
				 "\n"
				 "    <configfile>                Location of ORCA configuration file.\n"
				 "\n",
				 progname);
	 }

   exit (EXIT_FAILURE);
}

static void parse_options (options_t *opts,int argc,char *argv[])
{
   const char *progname;
   int i;

   (progname = strrchr (argv[0],'/')) ? progname++ : (progname = argv[0]);

   /* is the user requesting help? */
   if (argc == 2 && (!strcmp (argv[1],"-h") || !strcmp (argv[1],"--help")))
	 help (progname);

   if (argc < 2) usage (progname);

   opts->logfile = NULL;
   opts->loglevel = -1;
   opts->daemon = 0;

   for (i = 1; i < argc - 1; i++)
	 {
		if (!strcmp (argv[i],"-d") || !strcmp (argv[i],"--daemon"))
		  opts->daemon = 1;
		else if (!strcmp (argv[i],"-f") || !strncmp (argv[i],"--logfile=",10))
		  {
			 char *arg = strlen (argv[i]) == 2 ? argv[++i] : strchr (argv[i],'=') + 1;
			 if (i == argc - 1 || *arg == '\0') usage (progname);
			 opts->logfile = arg;
		  }
		else if (!strcmp (argv[i],"-l") || !strncmp (argv[i],"--loglevel=",11))
		  {
			 char *arg = strlen (argv[i]) == 2 ? argv[++i] : strchr (argv[i],'=') + 1;
			 if (i == argc - 1 || *arg == '\0') usage (progname);

			 if (!strcmp (arg,"quiet"))
			   opts->loglevel = LOG_QUIET;
			 else if (!strcmp (arg,"errors"))
			   opts->loglevel = LOG_ERROR;
			 else if (!strcmp (arg,"warnings"))
			   opts->loglevel = LOG_WARNING;
			 else if (!strcmp (arg,"normal"))
			   opts->loglevel = LOG_NORMAL;
			 else if (!strcmp (arg,"verbose"))
			   opts->loglevel = LOG_VERBOSE;
			 else if (!strcmp (arg,"debug"))
			   opts->loglevel = LOG_DEBUG;
			 else if (!strcmp (arg,"noisy"))
			   opts->loglevel = LOG_NOISY;
			 else usage (progname);
		  }
		else usage (progname);
	 }

   opts->configfile = argv[i];

   if (opts->daemon && opts->logfile == NULL)
	 usage (progname);

   if (opts->loglevel == -1)
	 opts->loglevel = opts->logfile == NULL ? LOG_NOISY : LOG_VERBOSE;
}

#ifdef DEBUG
static void print_settings (const options_t *o,const config_t *c)
{
   int i;
   const char *str[] = { "quiet", "errors", "warnings", "normal", "verbose", "debug", "noisy" };

   log_printf (LOG_DEBUG,
			   "[options]\n"
			   "   logfile: %s\n"
			   "  loglevel: %s\n"
			   "    daemon: %s\n"
			   "configfile: %s\n"
			   "\n",
			   o->logfile,
			   str[o->loglevel],
			   (o->daemon ? "yes" : "no"),
			   o->configfile);

   log_printf (LOG_DEBUG,
			   "[config]\n"
			   "community: %s\n"
			   "  modules: %s\n"
			   "  confdir: %s\n"
			   "   listen: %d.%d.%d.%d:%d\n"
			   "    allow:",
			   c->community,
			   c->modules,
			   c->confdir,
			   ADDR(c->addr.sin_addr.s_addr),ntohs (c->addr.sin_port));

   for (i = 0; i < c->n; i++)
	 {
		if (i) log_putc (LOG_DEBUG,',');
		log_printf (LOG_DEBUG," %d.%d.%d.%d/%d.%d.%d.%d",ADDR(c->allow[i].address),ADDR(c->allow[i].netmask));
	 }

   log_printf (LOG_DEBUG,
			   "\n"
			   "reuseaddr: %s\n"
			   "      uid: %u\n"
			   "      gid: %u\n"
			   "  pidfile: %s\n"
			   "\n",
			   c->reuseaddr ? "yes" : "no",
			   c->uid,
			   c->gid,
			   c->pidfile);
}
#endif	/* #ifdef DEBUG */

static int create_pidfile (const char *filename)
{
   int fd,result;
   char str[32];

   if ((fd = creat (filename,0644)) < 0) return (-1);

   sprintf (str,"%u\n",getpid ());

   if ((result = write (fd,str,strlen (str))) != strlen (str))
	 {
		int saved = result < 0 ? errno : EIO;
		close (fd);
		errno = saved;
		return (-1);
	 }

   close (fd);

   return (0);
}

int main (int argc,char *argv[])
{
   options_t opts;
   const config_t *conf;
   ber_t packet;
   int fd,result;
   fd_set rfds;
   struct timeval tv;
   int flags = LOG_DETECT_DUPLICATES | LOG_DETECT_FLOODING;

   /* ignore the important signals */
   signal (SIGHUP,SIG_IGN);
   signal (SIGINT,SIG_IGN);
   signal (SIGTERM,SIG_IGN);
   signal (SIGUSR1,SIG_IGN);
   signal (SIGUSR2,SIG_IGN);
   signal (SIGTSTP,SIG_IGN);

   /* parse command-line options */
   parse_options (&opts,argc,argv);

   /* initialize memory detection and logging subsystems */
   mem_open (NULL);
   if (log_open (opts.logfile,opts.loglevel,flags + (opts.logfile == NULL ? LOG_HAVE_COLORS : LOG_DEBUG_PREFIX_ONLY)) < 0)
	 {
		mem_close ();
		fprintf (stderr,"unable to initialize logging system: %s\n",strerror (errno));
		exit (EXIT_FAILURE);
	 }
   atexit (mem_close);
   atexit (log_close);

   /* parse configuration file */
   if ((conf = config_parse (opts.configfile)) == NULL)
	 exit (EXIT_FAILURE);

#ifdef DEBUG
   /* print configuration and command-line options */
   print_settings (&opts,conf);
#endif	/* #ifdef DEBUG */

   /* create a socket for communication */
   if ((fd = socket (PF_INET,SOCK_DGRAM,IPPROTO_UDP)) < 0)
	 {
		log_printf (LOG_ERROR,"unable to create socket: %s\n",strerror (errno));
		exit (EXIT_FAILURE);
	 }

   /* if the user wants to reuse ports, try to set SO_REUSEADDR */
   flags = 1;
   if (conf->reuseaddr && setsockopt (fd,SOL_SOCKET,SO_REUSEADDR,&flags,sizeof (flags)) < 0)
	 log_printf (LOG_WARNING,"setsockopt SO_REUSEADDR failed (%s). not reusing local addresses\n",strerror (errno));

   /* bind to the socket */
   if (bind (fd,(struct sockaddr *) &conf->addr,sizeof (struct sockaddr)) < 0)
	 {
		log_printf (LOG_ERROR,"couldn't bind to socket %d (%d.%d.%d.%d:%d): %s\n",
					fd,
					ADDR (conf->addr.sin_addr.s_addr),
					conf->addr.sin_port,
					strerror (errno));
		close (fd);
		exit (EXIT_FAILURE);
	 }

   /* select non-blocking I/O */
   if ((flags = fcntl (fd,F_GETFL)) < 0 || fcntl (fd,F_SETFL,flags | O_NONBLOCK) < 0)
	 {
		log_printf (LOG_ERROR,"unable to select non-blocking I/O: %s\n",strerror (errno));
		close (fd);
		exit (EXIT_FAILURE);
	 }

   if (opts.daemon)
	 {
		/* fork and become a daemon */
		if (daemon (0,0) < 0)
		  {
			 log_printf (LOG_ERROR,"unable to daemonize myself: %s\n",strerror (errno));
			 close (fd);
			 exit (EXIT_FAILURE);
		  }

		/* now create the pidfile */
		if (create_pidfile (conf->pidfile) < 0)
		  {
			 log_printf (LOG_ERROR,"unable to create pid file %s: %s\n",conf->pidfile,strerror (errno));
			 close (fd);
			 exit (EXIT_FAILURE);
		  }
	 }

   /* change ownership of log file */
   if (opts.logfile != NULL && chown (opts.logfile,conf->uid,conf->gid) < 0)
	 {
		log_printf (LOG_ERROR,"unable to change ownership of log file %s to uid %u, gid %u: %s\n",
					opts.logfile,conf->uid,conf->gid,strerror (errno));
		close (fd);
		exit (EXIT_FAILURE);
	 }

   /* set process gid */
   if (setregid (conf->gid,conf->gid) < 0)
	 {
		log_printf (LOG_ERROR,"unable to change process group id to %u\n",conf->gid);
		close (fd);
		exit (EXIT_FAILURE);
	 }

   /* set process uid */
   if (setreuid (conf->uid,conf->uid) < 0)
	 {
		log_printf (LOG_ERROR,"unable to change process user id to %u\n",conf->uid);
		close (fd);
		exit (EXIT_FAILURE);
	 }

   /* now that we've given up our priviledges, load all agent loadable modules */
   if (mod_open (conf->modules,conf->confdir) < 0)
	 {
		close (fd);
		exit (EXIT_FAILURE);
	 }

   /* create input buffer. we use the extra byte to detect truncated packets. */
   if ((packet.buf = (uint8_t *) mem_alloc (UDP_DATAGRAM_SIZE + 1)) == NULL)
	 {
		log_printf (LOG_ERROR,"mem_alloc failed: %s\n",strerror (errno));
		close (fd);
		exit (EXIT_FAILURE);
	 }

   /* catch all the signals we're interested in */
   signal (SIGINT,catch_signals);
   signal (SIGTERM,catch_signals);
   signal (SIGUSR1,catch_signals);

   /* once we enter the main loop, we try our best not to abort */
   while (!finished)
	 {
		/* reset the log file */
		if (rotated)
		  {
			 if (log_reset () < 0) break;
			 rotated = 0;
		  }

		/* listen for incoming connections */
		FD_ZERO (&rfds);
		FD_SET (fd,&rfds);

		/* wait 5 minutes */
		tv.tv_sec = 300;
		tv.tv_usec = 0;

		result = select (fd + 1,&rfds,NULL,NULL,&tv);

		/* did something bad happen? */
		if (result < 0 && errno != EINTR && errno != ENOMEM)
		  {
			 log_printf (LOG_ERROR,"select failed: %s\n",strerror (errno));
			 break;
		  }

		/* time to do some work */
		if (result > 0)
		  {
			 packet.offset = 0;
			 packet.size = UDP_DATAGRAM_SIZE;
			 if (FD_ISSET (fd,&rfds)) process_incoming (fd,&packet,conf);
		  }
	 }

   /* clean up */
   mem_free (packet.buf);
   close (fd);

   /* did something go wrong? */
   if (!finished) exit (EXIT_FAILURE);

   exit (EXIT_SUCCESS);
}

