
/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <errno.h>

#include <debug/memory.h>
#include <ber/ber.h>

#include <tinysnmp/tinysnmp.h>
#include <tinysnmp/agent/odb.h>
#include <tinysnmp/agent/mod.h>
#include <tinysnmp/compat/inttypes.h>

#include "snmp.h"
#include "mod.h"

static int encode_type (ber_t *e,const snmp_value_t *value,uint8_t *status,uint32_t *index,int n)
{
   switch (value->type)
	 {
	  case BER_INTEGER:			return (ber_encode_integer (e,value->data.INTEGER));
	  case BER_Counter32:		return (ber_encode_counter32 (e,value->data.Counter32));
	  case BER_Gauge32:			return (ber_encode_gauge32 (e,value->data.Gauge32));
	  case BER_TimeTicks:		return (ber_encode_timeticks (e,value->data.TimeTicks));
	  case BER_Counter64:		return (ber_encode_counter64 (e,value->data.Counter64));
	  case BER_OID:				return (ber_encode_oid (e,value->data.OID));
	  case BER_OCTET_STRING:	return (ber_encode_octet_string (e,&value->data.OCTET_STRING));
	  case BER_IpAddress:		return (ber_encode_ipaddress (e,value->data.IpAddress));
	  default:
		if (status == noError)
		  {
			 *status = noSuchName;
			 *index = n;
		  }
		return (ber_encode_null (e));
	 }
}

static int encode_value (ber_t *e,uint32_t offset,const uint32_t *oid,uint8_t *status,uint32_t *index,int n)
{
   const snmp_value_t *value;

   if (mod_update (oid) < 0)
	 {
		*status = genErr;
		*index = n;
		return (-1);
	 }

   if ((value = odb_find (oid)) == NULL)
	 {
		*status = noSuchName;
		*index = n;
		if (ber_encode_null (e) < 0) return (-1);					/* value NULL						*/
	 }
   else if (encode_type (e,value,status,index,n) < 0)				/* value							*/
	 return (-1);

   if (ber_encode_oid (e,oid) < 0 ||								/* name OBJECT IDENTIFIER			*/
	   ber_encode_sequence (e,e->offset - offset) < 0)				/* VarBind ::= SEQUENCE				*/
	 return (-1);

   return (0);
}

static int encode_next_value (ber_t *e,uint32_t offset,const uint32_t *oid,uint8_t *status,uint32_t *index,int n)
{
   snmp_next_value_t *next;

   if (mod_update_next (oid) < 0)
	 {
		*status = genErr;
		*index = n;
		return (-1);
	 }

   if ((next = odb_find_next (oid)) == NULL)
	 {
		*status = noSuchName;
		*index = n;
		if (oid != NULL)
		  {
			 if (ber_encode_null (e) < 0 ||							/* value NULL						*/
				 ber_encode_oid (e,oid) < 0 ||						/* name OBJECT IDENTIFIER			*/
				 ber_encode_sequence (e,e->offset - offset) < 0)	/* VarBind ::= SEQUENCE				*/
			   return (-1);
		  }
	 }
   else
	 {
		int result;

		result = encode_type (e,&next->value,status,index,n);		/* value							*/

		if (next->value.type == BER_OID)
		  mem_free (next->value.data.OID);
		else if (next->value.type == BER_OCTET_STRING && next->value.data.OCTET_STRING.len)
		  mem_free (next->value.data.OCTET_STRING.buf);

		if (!result) result = ber_encode_oid (e,next->oid);			/* name OBJECT IDENTIFIER			*/

		mem_free (next->oid);
		mem_free (next);

		return (ber_encode_sequence (e,e->offset - offset));		/* VarBind ::= SEQUENCE				*/
	 }

   return (0);
}

/*
 * Encode a GetResponse-PDU packet.
 *
 *      e           the buffer (see BER library)
 *      pdu         structure containing all the PDU data that should be encoded
 *      status      Error Status (see RFC 1157)
 *      index       Error Index (see RFC 1157)
 *
 * Returns 0 if successful, -1 if the message couldn't fit into the buffer or
 * if any of the ObjectID modules could not be updated.
 */
int encode_get_response (ber_t *e,const snmp_pdu_t *pdu,uint8_t status,uint32_t index)
{
   int i;
   uint32_t offset;

   mod_clear ();

   if (status)
	 for (i = pdu->n - 1; i >= 0; i--)
	   {
		  offset = e->offset;

		  if (ber_encode_null (e) < 0 ||							/* value NULL						*/
			  ber_encode_oid (e,pdu->oid[i]) < 0 ||					/* name OBJECT IDENTIFIER			*/
			  ber_encode_sequence (e,e->offset - offset) < 0)		/* VarBind ::= SEQUENCE				*/
			return (-1);
	   }
   else if (pdu->type != BER_GetNextRequest || pdu->n)
	 {
		for (i = pdu->n - 1; i >= 0; i--)
		  {
			 offset = e->offset;

			 /* value, name, VarBind */
			 if (pdu->type == BER_GetRequest)
			   {
				  if (encode_value (e,offset,pdu->oid[i],&status,&index,i + 1) < 0)
					return (-1);
			   }
			 else if (pdu->type == BER_GetNextRequest)
			   {
				  if (encode_next_value (e,offset,pdu->oid[i],&status,&index,i + 1) < 0)
					return (-1);
			   }
		  }
	 }
   else
	 {
		/* value, name, VarBind */
		if (encode_next_value (e,e->offset,NULL,&status,&index,0) < 0)
		  return (-1);
	 }

   if (ber_encode_sequence (e,e->offset) < 0 ||						/* VarBindList ::= SEQUENCE OF		*/
	   ber_encode_integer (e,index) < 0 ||							/* ErrorIndex INTEGER				*/
	   ber_encode_integer (e,status) < 0 ||							/* ErrorStatus INTEGER				*/
	   ber_encode_integer (e,pdu->RequestID) < 0 ||					/* Request-ID INTEGER				*/
	   ber_encode_get_response (e) < 0 ||							/* PDU { GetResponse-PDU(2) }		*/
	   ber_encode_octet_string (e,&pdu->community) ||				/* community OCTET STRING			*/
	   ber_encode_integer (e,pdu->version) < 0 ||					/* version INTEGER { version-1(0) }	*/
	   ber_encode_sequence (e,e->offset) < 0)						/* Message ::= SEQUENCE				*/
	 return (-1);

   return (0);
}

/*
 * Decode GetRequest-PDU or GetNextRequest-PDU.
 *
 *      d           the buffer (see BER library)
 *      pdu         structure containing all the PDU data that were decoded
 *
 * Returns 0 if successful, -1 if some error occurred. Check errno to see
 * what error occurred. The following errors is possible:
 *
 *      ENOMEM      Out of memory
 *      EINVAL      The packet is a valid SNMP packet, but either some of the error
 *                  flags is set or it is an unsupported SNMP version or we were
 *                  unable to decode a tag properly
 */
int decode_pdu (snmp_pdu_t *pdu,ber_t *d)
{
   int32_t ErrorStatus,ErrorIndex;
   uint32_t *oid,**ptr,offset;
   int i,result;

   pdu->community.len = 0;
   if ((result = ber_decode_sequence (d)) < 0 ||						/* Message ::= SEQUENCE					*/
	   (result = ber_decode_integer (&pdu->version,d)) < 0 ||			/* version INTEGER { version-1(0) }		*/
	   (result = ber_decode_octet_string (&pdu->community,d)) < 0)		/* community OCTET STRING				*/
	 {
		if (pdu->community.len) mem_free (pdu->community.buf);
		errno = -result;
		return (-1);
	 }

   offset = d->offset;
   if ((result = ber_decode_get_request (d)) < 0)						/* PDU { GetRequest-PDU(0) }			*/
	 {
		d->offset = offset;
		if ((result = ber_decode_get_next_request (d)) < 0)				/* PDU { GetNextRequest-PDU(1) }		*/
		  {
			 if (pdu->community.len) mem_free (pdu->community.buf);
			 errno = -result;
			 return (-1);
		  }
		pdu->type = BER_GetNextRequest;
	 }
   else pdu->type = BER_GetRequest;

   if ((result = ber_decode_integer (&pdu->RequestID,d)) < 0 ||			/* Request-ID INTEGER					*/
	   (result = ber_decode_integer (&ErrorStatus,d)) < 0 ||			/* ErrorStatus INTEGER { noError(0) }	*/
	   (result = ber_decode_integer (&ErrorIndex,d)) < 0 ||				/* ErrorIndex INTEGER					*/
	   (result = ber_decode_sequence (d)) < 0)							/* VarBindList ::= SEQUENCE OF			*/
	 {
		if (pdu->community.len) mem_free (pdu->community.buf);
		errno = -result;
		return (-1);
	 }

   if (pdu->version != SNMP_VERSION_1 || ErrorStatus || ErrorIndex)
	 {
		if (pdu->community.len) mem_free (pdu->community.buf);
		errno = EINVAL;
		return (-1);
	 }

   pdu->n = 0;
   pdu->oid = NULL;

   for (i = 0; d->offset < d->size; i++)
	 {
		if ((result = ber_decode_sequence (d)) < 0 ||			/* VarBind ::= SEQUENCE		*/
			(result = ber_decode_oid (&oid,d)) < 0 ||			/* name OBJECT IDENTIFIER	*/
			(result = ber_decode_null (d)) < 0)					/* value NULL				*/
		  {
			 if (pdu->community.len) mem_free (pdu->community.buf);
			 if (pdu->n)
			   {
				  for (i = 0; i < pdu->n; i++) mem_free (pdu->oid[i]);
				  mem_free (pdu->oid);
			   }
			 if (oid != NULL) mem_free (oid);
			 errno = -result;
			 return (-1);
		  }

		if ((ptr = (uint32_t **) mem_realloc (pdu->oid,(pdu->n + 1) * sizeof (uint32_t *))) == NULL)
		  {
			 if (pdu->community.len) mem_free (pdu->community.buf);
			 for (i = 0; i < pdu->n; i++) mem_free (pdu->oid[i]);
			 if (pdu->n) mem_free (pdu->oid);
			 if (oid != NULL) mem_free (oid);
			 errno = ENOMEM;
			 return (-1);
		  }

		pdu->oid = ptr;
		pdu->oid[i] = oid;
		pdu->n++;
	 }

   if ((pdu->type == BER_GetRequest && !pdu->n) || d->offset != d->size)
	 {
		if (pdu->community.len) mem_free (pdu->community.buf);
		if (pdu->n)
		  {
			 for (i = 0; i < pdu->n; i++) mem_free (pdu->oid[i]);
			 mem_free (pdu->oid);
		  }
		errno = EINVAL;
		return (-1);
	 }

   return (0);
}

