
/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <errno.h>

#include <debug/memory.h>
#include <debug/log.h>

#include <tinysnmp/compat/inttypes.h>

#include "tree.h"

#define DEPTH 32

typedef struct
{
   uint32_t n;
   uint32_t *branch;
   void *value;
} node_t;

typedef struct
{
   const char *filename;
   int line;
   const char *function;
   int level;
   int *prev;
   size_t n;
} show_priv_t;

/*
 * Destroy a tree. When this function returns, the tree is
 * in exactly the same state as when initialized.
 */
void tree_destroy (tree_t **tree,void (*purge)(void *))
{
   if (*tree != NULL)
	 {
		if ((*tree)->sibling != NULL) tree_destroy (&(*tree)->sibling,purge);
		if ((*tree)->child != NULL) tree_destroy (&(*tree)->child,purge);
		if ((*tree)->type == TREE_VALUE) purge ((*tree)->data.value);
		mem_free (*tree);
		*tree = NULL;
	 }
}

/* add_sibling_stub() and add_child_stub() depends on this */
static int add_stub (const char *filename,int line,const char *function,tree_t **tree,const node_t *node);

static int add_sibling_stub (const char *filename,int line,const char *function,tree_t **tree,const node_t *node)
{
   if (add_stub (filename,line,function,&(*tree)->sibling,node) < 0)
	 {
		if ((*tree)->sibling != NULL && (*tree)->sibling->child == NULL)
		  {
			 tree_t *sibling = (*tree)->sibling;
			 (*tree)->sibling = (*tree)->sibling->sibling;
			 mem_free (sibling);
		  }

		return (-1);
	 }

   if ((*tree)->sibling != NULL) (*tree)->sibling->parent = (*tree)->parent;

   return (0);
}

static int add_child_stub (const char *filename,int line,const char *function,tree_t **tree,const node_t *node)
{
   node_t next = { n: node->n - 1, branch: node->branch + 1, value: node->value };

   if (add_stub (filename,line,function,&(*tree)->child,&next) < 0)
	 {
		if ((*tree)->child == NULL && (*tree)->sibling == NULL)
		  {
			 mem_free (*tree);
			 *tree = NULL;
		  }

		return (-1);
	 }

   if ((*tree)->child != NULL) (*tree)->child->parent = *tree;

   return (0);
}

static tree_t *create_stub (const char *filename,int line,const char *function)
{
   tree_t *tree;

   if ((tree = (tree_t *) mem_alloc_stub (sizeof (tree_t),filename,line,function)) == NULL)
	 {
		errno = ENOMEM;
		return (NULL);
	 }

   tree->parent = tree->sibling = tree->child = NULL;

   return (tree);
}

static int add_stub (const char *filename,int line,const char *function,tree_t **tree,const node_t *node)
{
   if (!node->n)
	 {
		if (*tree != NULL)
		  {
			 errno = EEXIST;
			 return (-1);
		  }

		if ((*tree = create_stub (filename,line,function)) == NULL)
		  return (-1);

		(*tree)->type = TREE_VALUE;
		(*tree)->data.value = node->value;

		return (0);
	 }

   if (*tree == NULL)
	 {
		if ((*tree = create_stub (filename,line,function)) == NULL)
		  return (-1);

		(*tree)->type = TREE_NODE;
		(*tree)->data.node = node->branch[0];

		return (add_child_stub (filename,line,function,tree,node));
	 }

   if ((*tree)->type == TREE_NODE)
	 {
		if ((*tree)->data.node > node->branch[0])
		  {
			 tree_t *tmp;

			 if ((tmp = create_stub (filename,line,function)) == NULL)
			   return (-1);

			 tmp->type = TREE_NODE;
			 tmp->data.node = node->branch[0];
			 tmp->sibling = *tree;
			 tmp->parent = (*tree)->parent;

			 *tree = tmp;

			 return (add_child_stub (filename,line,function,tree,node));
		  }

		if ((*tree)->data.node < node->branch[0])
		  return (add_sibling_stub (filename,line,function,tree,node));

		if (node->n > 1)
		  return (add_child_stub (filename,line,function,tree,node));
	 }

   errno = EEXIST;
   return (-1);
}

/*
 * Add a new branch to the tree. Returns 0 if successful, -1
 * otherwise. Check errno to see what error occurred. The
 * following errors is possible:
 *
 *       ENOMEM     Out of memory
 *       EEXIST     A branch like this (or superset of) already exist
 */
int tree_add_stub (const char *filename,int line,const char *function,tree_t **tree,const branch_t *branch)
{
   node_t node = { n: branch->branch[0], branch: branch->branch + 1, value: branch->value };
   return (add_stub (filename,line,function,tree,&node));
}

static void *update (tree_t *tree,node_t *node)
{
   node_t next;

   if (tree == NULL || tree->type != TREE_NODE || !node->n || tree->data.node > node->branch[0] || tree->child == NULL)
	 return (NULL);

   if (tree->data.node < node->branch[0])
	 return (update (tree->sibling,node));

   if (tree->child->type == TREE_VALUE && node->n == 1)
	 {
		void *old = tree->child->data.value;
		tree->child->data.value = node->value;
		return (old);
	 }

   next.branch = node->branch + 1;
   next.n = node->n - 1;
   next.value = node->value;

   return (update (tree->child,&next));
}

/*
 * Update the value of an existing branch. Returns the old value
 * of the branch if successful, or NULL of the branch doesn't
 * exist.
 */
extern void *tree_update (tree_t *tree,branch_t *branch)
{
   node_t node = { n: branch->branch[0], branch: branch->branch + 1, value: branch->value };
   return (update (tree,&node));
}

static void *find (const tree_t *tree,node_t *node)
{
   node_t next;

   if (tree == NULL || tree->type != TREE_NODE || !node->n || tree->data.node > node->branch[0] || tree->child == NULL)
	 return (NULL);

   if (tree->data.node < node->branch[0])
	 return (find (tree->sibling,node));

   if (tree->child->type == TREE_VALUE && node->n == 1)
	 return (tree->child->data.value);

   next.branch = node->branch + 1;
   next.n = node->n - 1;

   return (find (tree->child,&next));
}

/*
 * Find a branch in a tree. Return the value of an existing
 * branch in the tree if successful, or NULL if the branch
 * doesn't exist.
 */
const void *tree_find (const tree_t *tree,const uint32_t *branch)
{
   node_t node = { n: branch[0], branch: (uint32_t *) branch + 1 };
   return (find (tree,&node));
}

static branch_t *find_first (const tree_t *tree)
{
   branch_t *result;
   const tree_t *tmp;
   uint32_t i,n;

   if ((result = (branch_t *) mem_alloc (sizeof (branch_t))) == NULL)
	 {
		errno = ENOMEM;
		return (NULL);
	 }

   n = 1;

   while (tree->child->type != TREE_VALUE) tree = tree->child;
   for (tmp = tree->parent; tmp != NULL; tmp = tmp->parent) n++;

   if ((result->branch = (int *) mem_alloc ((n + 1) * sizeof (uint32_t))) == NULL)
	 {
		mem_free (result);
		errno = ENOMEM;
		return (NULL);
	 }

   result->branch[0] = n;

   for (tmp = tree, i = n; tmp != NULL; tmp = tmp->parent)
	 result->branch[i--] = tmp->data.node;

   result->value = tree->child->data.value;

   return (result);
}

static branch_t *find_next (const tree_t *tree,const node_t *node)
{
   if (tree == NULL || tree->child == NULL || tree->type == TREE_VALUE || !node->n)
	 return (NULL);

   if (tree->data.node > node->branch[0])
	 return (find_first (tree));

   if (tree->data.node == node->branch[0])
	 {
		if (node->n != 1)
		  {
			 branch_t *result;
			 node_t next = { n: node->n - 1, branch: node->branch + 1 };

			 if ((result = find_next (tree->child,&next)) != NULL)
			   return (result);
		  }
		else if (tree->child->child != NULL)
		  return (find_first (tree));
	 }

   return (find_next (tree->sibling,node));
}

/*
 * Find the branch in the tree that lexographically succeeds
 * the specified one or NULL if one doesn't exist. The specified
 * branch doesn't need to exist in the tree and can also be a
 * partial branch. Check errno to see what error occurred. The
 * following errors is possible:
 *
 *       ENOMEM     Out of memory
 *       ENOENT     Specified branch does not lexographically
 *                  preceed an existing branch
 */
branch_t *tree_find_next (const tree_t *tree,const uint32_t *branch)
{
   node_t node = { n: branch[0], branch: (uint32_t *) branch + 1 };
   return (find_next (tree,&node));
}

static int _remove (tree_t **tree,node_t *node,void (*purge)(void *))
{
   if (*tree == NULL || (*tree)->child == NULL || (*tree)->type != TREE_NODE || !node->n || (*tree)->data.node > node->branch[0])
	 return (-1);

   if ((*tree)->data.node < node->branch[0])
	 {
		if ((*tree)->sibling != NULL)
		  {
			 tree_t *sibling = (*tree)->sibling->sibling;
			 if (_remove (&(*tree)->sibling,node,purge) < 0) return (-1);
			 if ((*tree)->sibling == NULL) (*tree)->sibling = sibling;
			 return (0);
		  }

		return (-1);
	 }

   if (node->n == 1)
	 {
		tree_t *tmp = (*tree)->sibling;
		(*tree)->sibling = NULL;
		tree_destroy (tree,purge);
		*tree = tmp;
		return (0);
	 }

   if ((*tree)->child->type == TREE_NODE)
	 {
		node_t next = { n: node->n - 1, branch: node->branch + 1 };
		tree_t *child = (*tree)->child->sibling;

		if (_remove (&(*tree)->child,&next,purge) < 0) return (-1);
		if ((*tree)->child == NULL) (*tree)->child = child;
		if ((*tree)->child == NULL)
		  {
			 tree_t *tmp = (*tree)->sibling;
			 mem_free (*tree);
			 *tree = tmp != NULL ? tmp : NULL;
		  }

		return (0);
	 }

   return (-1);
}

/*
 * Remove a branch (or part thereof) from a tree. Returns 0 if
 * successful, or -1 if the branch subtree doesn't exist.
 */
int tree_remove (tree_t **tree,const uint32_t *branch,void (*purge)(void *))
{
   node_t node = { n: branch[0], branch: (uint32_t *) branch + 1 };
   return (_remove (tree,&node,purge));
}

#define priv_printf(fmt,args...) log_printf_stub(priv->filename,priv->line,priv->function,priv->level,fmt,##args)
#define priv_puts(s) log_puts_stub(priv->filename,priv->line,priv->function,priv->level,s)
#define priv_putc(c) log_putc_stub(priv->filename,priv->line,priv->function,priv->level,c)

static int show_n (show_priv_t *priv,int depth,const tree_t *tree,show_func_t show_stub)
{
   int i;

   if (tree == NULL || tree->child == NULL || tree->type != TREE_NODE)
	 {
		errno = EINVAL;
		return (-1);
	 }

   for (i = 0; i < depth; i++)
	 if (priv_printf ("  %c",priv->prev[i]) < 0)
	   return (-1);

   if (priv_printf ("  %c %" PRIu32,tree->sibling != NULL ? '' : '',tree->data.node) < 0)
	 return (-1);

   if (tree->child->child == NULL)
	 {
		if (tree->child->type != TREE_VALUE)
		  {
			 errno = EINVAL;
			 return (-1);
		  }

		priv_puts (" :: ");
		if (show_stub (priv->filename,priv->line,priv->function,priv->level,tree->child->data.value) < 0)
		  return (-1);
	 }

   priv_putc ('\n');

   if (depth >= priv->n)
	 {
		int *ptr;
		priv->n += DEPTH;
		if ((ptr = (int *) mem_realloc (priv->prev,priv->n * sizeof (int))) == NULL)
		  return (-1);
		priv->prev = ptr;
	 }
   priv->prev[depth] = tree->sibling != NULL ? '' : ' ';

   if (tree->child->child != NULL)
	 if (show_n (priv,depth + 1,tree->child,show_stub) < 0)
	   return (-1);

   if (tree->sibling != NULL)
	 if (show_n (priv,depth,tree->sibling,show_stub) < 0)
	   return (-1);

   return (0);
}

#undef priv_putc
#undef priv_puts
#undef priv_printf

/*
 * Display a tree. Returns 0 if successful, or -1 if some error
 * occurred. Check errno to see what error occurred.
 *
 * The show_stub function should return 0 if successful, or -1 if
 * some error occurred. If an error did occur, the show function
 * should set errno.
 *
 * The show_stub function prototype is as follows:
 *
 *     int show_stub(const char *filename,
 *                   int line,
 *                   const char *function,
 *                   const void *value);
 */
int tree_show_stub (const char *filename,int line,const char *function,int level,const tree_t *tree,show_func_t show_stub)
{
   show_priv_t priv;

   if (tree == NULL)
	 return (0);

   priv.filename = filename;
   priv.line = line;
   priv.function = function;
   priv.level = level;
   priv.n = DEPTH;

   if ((priv.prev = (int *) mem_alloc (priv.n * sizeof (int))) == NULL)
	 return (-1);

   if (show_n (&priv,0,tree,show_stub) < 0)
	 {
		int saved = errno;
		mem_free (priv.prev);
		errno = saved;
		return (-1);
	 }

   mem_free (priv.prev);

   return (0);
}

