
/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stddef.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>

#include <tinysnmp/tinysnmp.h>
#include <tinysnmp/agent/mod.h>
#include <tinysnmp/agent/odb.h>

#include <ber/ber.h>
#include <debug/log.h>

#include "meminfo.h"

static int memory_update (const struct mod_r *mod)
{
   uint32_t oid[13] = { 12, 43, 6, 1, 4, 1, 10002, 1, 1, 1, 1, 0, 0 };
   snmp_value_t value;
   struct meminfo info;

   odb_remove (mod->oid);

   if (getmeminfo (&info))
	 {
		log_printf (LOG_ERROR,"%s: failed to get memory resource information\n",mod->name);
		return (-1);
	 }

   value.type = BER_Gauge32;

   oid[11] = 1;
   value.data.Gauge32 = info.mem_total / 1024;
   if (odb_add (oid,&value))
	 {
		log_printf (LOG_ERROR,
					"%s: couldn't add memory.memTotal.0: %s\n",
					mod->name,strerror (errno));
		return (-1);
	 }

   oid[11] = 2;
   value.data.Gauge32 = info.mem_free / 1024;
   if (odb_add (oid,&value))
	 {
		log_printf (LOG_ERROR,
					"%s: couldn't add memory.memFree.0: %s\n",
					mod->name,strerror (errno));
		return (-1);
	 }

   oid[11] = 3;
   value.data.Gauge32 = info.mem_buffer / 1024;
   if (odb_add (oid,&value))
	 {
		log_printf (LOG_ERROR,
					"%s: couldn't add memory.memBuffer.0: %s\n",
					mod->name,strerror (errno));
		return (-1);
	 }

   oid[11] = 4;
   value.data.Gauge32 = info.mem_cache / 1024;
   if (odb_add (oid,&value))
	 {
		log_printf (LOG_ERROR,
					"%s: couldn't add memory.memCache.0: %s\n",
					mod->name,strerror (errno));
		return (-1);
	 }

   return (0);
}

static int swap_update (const struct mod_r *mod)
{
   uint32_t oid[13] = { 12, 43, 6, 1, 4, 1, 10002, 1, 1, 1, 2, 0, 0 };
   snmp_value_t value;
   struct swapinfo info;

   odb_remove (mod->oid);

   if (getswapinfo (&info))
	 {
		log_printf (LOG_ERROR,"%s: failed to get swap resource information\n",mod->name);
		return (-1);
	 }

   value.type = BER_Gauge32;

   oid[11] = 1;
   value.data.Gauge32 = info.swap_total / 1024;
   if (odb_add (oid,&value))
	 {
		log_printf (LOG_ERROR,
					"%s: couldn't add swap.swapTotal.0: %s\n",
					mod->name,strerror (errno));
		return (-1);
	 }

   oid[11] = 2;
   value.data.Gauge32 = info.swap_free / 1024;
   if (odb_add (oid,&value))
	 {
		log_printf (LOG_ERROR,
					"%s: couldn't add swap.swapFree.0: %s\n",
					mod->name,strerror (errno));
		return (-1);
	 }

   return (0);
}

/* iso.org.dod.internet.private.enterprises.frogfoot.servers.system.resources.memory */
static const uint32_t memory_oid[11] = { 10, 43, 6, 1, 4, 1, 10002, 1, 1, 1, 1 };

/* iso.org.dod.internet.private.enterprises.frogfoot.servers.system.resources.swap */
static const uint32_t swap_oid[11] = { 10, 43, 6, 1, 4, 1, 10002, 1, 1, 1, 2 };

const mod_t mod[] =
{
   { "resources/memory", NULL, memory_update, NULL, memory_oid },
   { "resources/swap", NULL, swap_update, NULL, swap_oid },
   { NULL, NULL, NULL, NULL, NULL }
};

