#ifndef LIB_IPC_H
#define LIB_IPC_H

/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <tinysnmp/compat/inttypes.h>
#include <abz/aton.h>

#define IPC_QUALITY_BRONZE		0x00
#define IPC_QUALITY_SILVER		0x01
#define IPC_QUALITY_GOLD		0x02
#define IPC_QUALITY_PLATINUM	0x03

typedef struct
{
   uint32_t id;							/* unique id associated with client					*/
   char *name;							/* client name										*/
   uint8_t quality;						/* service quality									*/
   uint32_t input;						/* expected incoming traffic (in bits per second)	*/
   uint32_t output;						/* expected outgoing traffic (in bits per second)	*/
   uint32_t n;							/* number of entries in list of networks			*/
   struct network *network;				/* list of client networks							*/
} ipc_client_info_t;

#define IPC_PROTO_UNKNOWN		0x00
#define IPC_PROTO_FTP			0x01
#define IPC_PROTO_SSH			0x02
#define IPC_PROTO_TELNET		0x03
#define IPC_PROTO_SMTP			0x04
#define IPC_PROTO_DNS			0x05
#define IPC_PROTO_TFTP			0x06
#define IPC_PROTO_HTTP			0x07
#define IPC_PROTO_POP3			0x08
#define IPC_PROTO_IMAP			0x09
#define IPC_PROTO_SNMP			0x0a
#define IPC_PROTO_IRC			0x0b
#define IPC_PROTO_TOTAL			0x0c

#define IPC_TYPE_RECEIVED		0x00
#define IPC_TYPE_TRANSMITTED	0x01

typedef struct
{
   uint8_t protocol;					/* protocol											*/
   uint8_t type;						/* type												*/
   uint64_t value;						/* counter value									*/
} ipc_counter_t;

typedef struct
{
   uint32_t n;							/* number of entries in list of counters			*/
   ipc_counter_t *counter;				/* list of counters									*/
} ipc_client_counters_t;

#define IPC_PACKET_CLIENTS		0x00
#define IPC_PACKET_INFO			0x01
#define IPC_PACKET_COUNTERS		0x02
#define IPC_PACKET_UPDATE		0x03
#define IPC_PACKET_UPTIME		0x04

#define IPC_PACKET_MASK			(IPC_PACKET_CLIENTS | IPC_PACKET_INFO | IPC_PACKET_COUNTERS | IPC_PACKET_UPDATE | IPC_PACKET_UPTIME)

typedef struct
{
   uint32_t n;							/* number of clients								*/
   ipc_client_info_t *info;				/* info structure for each client					*/
   ipc_client_counters_t *counter;		/* counter structure for each client				*/
   uint64_t uptime;						/* timestamp of when rules were last added			*/
} ipc_client_t;

typedef enum
{
   IPC_TRANSFER_READ = 0,
   IPC_TRANSFER_WRITE = 1
} ipc_transfer_t;

/*
 * transfer data between client and server.
 *
 *        level     log level
 *        fd        file descriptor of socket
 *        buf       pointer to buffer
 *        count     number of bytes to transfer
 *        type      transfer type (read/write)
 *
 * returns 0 if successful, -1 if some error occurred. errors are
 * logged automatically.
 */
#define ipc_transfer(level,fd,buf,count,type) ipc_transfer_stub(__FUNCTION__,level,fd,buf,count,type)
extern int ipc_transfer_stub (const char *function,int level,int fd,void *buf,ssize_t count,ipc_transfer_t type);

#endif	/* #ifndef LIB_IPC_H */
