#ifndef IP_H
#define IP_H

/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <tinysnmp/compat/inttypes.h>
#include <abz/aton.h>

#define MATCH_INCOMING	0
#define MATCH_OUTGOING	1

/*
 * Add rules to count traffic that matches the specified criteria. Returns 0
 * if successful, -1 otherwise. The function automatically logs errors.
 *
 *        level     Log level
 *        n         Rule number
 *        proto     IP protocol (IPPROTO_IP, IPPROTO_TCP, or IPPROTO_UDP)
 *        net       Source/Destination network to match
 *        port      Port number (if 0, any port matches)
 *        iface     Name of interface on which to match packets
 *        src       If 0, match destination, else match source
 */
extern int fw_add_counter (int level,uint16_t n,int proto,const struct network *net,uint16_t port,const char *iface,int src);

/*
 * Delete rules that match the specified rule number. Returns 0 if
 * successful, -1 otherwise. The function automatically logs errors.
 *
 *        level     Log level
 *        n         Rule number
 */
extern int fw_del_counters (int level,uint16_t n);

/*
 * Get a snapshot of the current firewall rules. This is needed by
 * fw_get_counters(). Returns a pointer to the allocated data structure
 * of NULL if some error occurred. The function automatically logs errors.
 *
 *        level     Log level
 */
extern void *fw_get_rules (int level);

/*
 * Return the sum of the interface byte count counters of the rules
 * that match the specified rule number. Returns 0 if one of more
 * rules matched, or -1 if no firewall rules match the specified
 * rule number.
 *
 *        rules     List of rules allocated by fw_get_rules()
 *        n         Rule number
 *        count     Sum of matched rule byte counters
 */
extern int fw_get_counters (void *rules,uint16_t n,uint64_t *count);

/*
 * Free memory allocated by the fw_get_rules() function.
 *
 *        rules     List of rules allocated by fw_get_rules()
 */
extern void fw_destroy_rules (void *rules);

#endif	/* #ifndef IP_H */
