
/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include <debug/log.h>
#include <debug/memory.h>

#include <tinysnmp/compat/inttypes.h>

#include "ipc.h"
#include "config.h"

static uint8_t protocol (protocol_t proto)
{
   switch (proto)
	 {
	  default:		return (IPC_PROTO_UNKNOWN);
	  case ftp:		return (IPC_PROTO_FTP);
	  case ssh:		return (IPC_PROTO_SSH);
	  case telnet:	return (IPC_PROTO_TELNET);
	  case smtp:	return (IPC_PROTO_SMTP);
	  case dns:		return (IPC_PROTO_DNS);
	  case tftp:	return (IPC_PROTO_TFTP);
	  case http:	return (IPC_PROTO_HTTP);
	  case pop3:	return (IPC_PROTO_POP3);
	  case imap:	return (IPC_PROTO_IMAP);
	  case snmp:	return (IPC_PROTO_SNMP);
	  case irc:		return (IPC_PROTO_IRC);
	  case total:	return (IPC_PROTO_TOTAL);
	 }
}

/*
 * receive a command from the client. the command is stored in ``cmd''
 * and if the retrieved command also includes a client index, said
 * index is stored in ``client''. returns 0 if successful, -1 if some
 * error occurred. the function automatically logs errors at the
 * LOG_WARNING level.
 */
int ipc_get (int fd,uint32_t *cmd,uint32_t *client)
{
   uint32_t buf[2];

   if (ipc_transfer (LOG_WARNING,fd,buf,8,IPC_TRANSFER_READ) < 0)
	 return (-1);

   if (buf[0] < 8 || (buf[1] & ~IPC_PACKET_MASK))
	 {
		log_printf (LOG_WARNING,"%s: invalid packet format. broken client?\n",__FUNCTION__);
		return (-1);
	 }

   *cmd = buf[1];

   if (*cmd == IPC_PACKET_UPDATE || *cmd == IPC_PACKET_CLIENTS || *cmd == IPC_PACKET_UPTIME)
	 return (buf[0] == 8 ? 0 : -1);

   if (buf[0] != 12)
	 return (-1);

   if (ipc_transfer (LOG_WARNING,fd,buf,4,IPC_TRANSFER_READ) < 0)
	 return (-1);

   *client = buf[0];

   return (0);
}

/*
 * send a timestamp of when the ruleset were added. returns 0 if
 * successful, -1 if some error occurred. the function automatically
 * logs errors at the LOG_WARNING level.
 */
int ipc_put_uptime (int fd,uint64_t timestamp)
{
   uint32_t buf[4];

   buf[0] = sizeof (buf);
   buf[1] = IPC_PACKET_UPTIME;
   ((uint64_t *) buf)[1] = timestamp;

   return (ipc_transfer (LOG_WARNING,fd,buf,sizeof (buf),IPC_TRANSFER_WRITE));
}

/*
 * send the number of clients to the client. returns 0 if successful,
 * -1 if some error occurred. the function automatically logs errors
 * at the LOG_WARNING level.
 */
int ipc_put_clients (int fd,uint32_t n)
{
   uint32_t buf[3] = { 12, IPC_PACKET_CLIENTS, n };
   return (ipc_transfer (LOG_WARNING,fd,buf,sizeof (buf),IPC_TRANSFER_WRITE));
}

static uint8_t *encode_64 (uint8_t **s,uint64_t value)
{
   *((uint64_t *) *s) = value;
   return (*s + 8);
}

static uint8_t *encode_32 (uint8_t **s,uint32_t value)
{
   *((uint32_t *) *s) = value;
   return (*s + 4);
}

static uint8_t *encode_8 (uint8_t **s,uint8_t value)
{
   **s = value;
   return (*s + 1);
}

static uint8_t *encode_name (uint8_t **s,const char *name)
{
   uint32_t i,len = strlen (name);
   *s = encode_32 (s,len);
   for (i = 0; i < len; i++) *s = encode_8 (s,name[i]);
   return (*s);
}

static uint8_t *encode_network (uint8_t **s,const struct network *network)
{
   *s = encode_32 (s,network->address);
   *s = encode_32 (s,network->netmask);
   return (*s);
}

/*
 * send the client information of specified client to the client (that
 * sounds confusing, doesn't it *g*). returns 0 if successful, -1 if
 * some error occurred. the function automatically logs errors at the
 * LOG_WARNING level.
 */
int ipc_put_info (int fd,const client_t *client)
{
   uint8_t *s,*buf;
   uint32_t i,length =
	 8 +								/* length, type of packet						*/
	 4 +								/* client id									*/
	 4 + strlen (client->info.name) +	/* client name (length + name)					*/
	 1 +								/* service quality								*/
	 8 +								/* incoming / ougoing traffic					*/
	 4 +								/* number of client networks					*/
	 8 * client->info.n;				/* client networks (n * (address + netmask))	*/
   int result;

   if ((buf = (uint8_t *) mem_alloc (length)) == NULL)
	 {
		log_printf (LOG_WARNING,"%s: unable to allocate memory for temporary storage: %s\n",__FUNCTION__,strerror (errno));
		return (-1);
	 }

   s = buf;

   s = encode_32 (&s,length);
   s = encode_32 (&s,IPC_PACKET_INFO);
   s = encode_32 (&s,client->info.id);
   s = encode_name (&s,client->info.name);
   s = encode_8 (&s,client->info.quality);
   s = encode_32 (&s,client->info.input);
   s = encode_32 (&s,client->info.output);
   s = encode_32 (&s,client->info.n);
   for (i = 0; i < client->info.n; i++)
	 s = encode_network (&s,client->info.networks + i);

   result = ipc_transfer (LOG_WARNING,fd,buf,length,IPC_TRANSFER_WRITE);

   mem_free (buf);

   return (result);
}

/*
 * send the client counters of specified client to the client. returns
 * 0 if successful, -1 if some error occurred. the function automatically
 * logs errors at the LOG_WARNING level.
 */
int ipc_put_counters (int fd,const client_t *client)
{
   uint8_t i,*s,*buf;
   uint32_t length = 12,counters = 0;
   int result;

   for (i = unknown; i < client->n; i++)
	 {
		if (client->recv[i].valid) length += 10, counters++;
		if (client->xmit[i].valid) length += 10, counters++;
	 }

   if ((buf = (uint8_t *) mem_alloc (length)) == NULL)
	 {
		log_printf (LOG_WARNING,"%s: unable to allocate memory for temporary storage: %s\n",__FUNCTION__,strerror (errno));
		return (-1);
	 }

   s = buf;

   s = encode_32 (&s,length);
   s = encode_32 (&s,IPC_PACKET_COUNTERS);
   s = encode_32 (&s,counters);

   for (i = unknown; i < client->n; i++)
	 {
		if (client->recv[i].valid)
		  {
			 s = encode_8 (&s,protocol (i));
			 s = encode_8 (&s,0);
			 s = encode_64 (&s,client->recv[i].counter);
		  }

		if (client->xmit[i].valid)
		  {
			 s = encode_8 (&s,protocol (i));
			 s = encode_8 (&s,1);
			 s = encode_64 (&s,client->xmit[i].counter);
		  }
	 }

   result = ipc_transfer (LOG_WARNING,fd,buf,length,IPC_TRANSFER_WRITE);

   mem_free (buf);

   return (result);
}

