
/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * CAVEATS:
 *
 * iptables doesn't have static rule numbers (i.e. if you insert an entry into
 * a chain, the rule numbers of all the other rules change), so we have a few
 * restrictions:
 *
 *   1) the rule number ``n'' passed for fw_add_counter() doesn't specify where
 *      in the chain we insert a rule, but is used as an index for our entry
 *      list (see next)
 *
 *   2) we keep a list of all added rules and then use this indexed list to
 *      find matching entries in the filter table. this of course is much
 *      slower than the ipfw-like way of searching for entries
 *
 * in addition, iptables do not have the "via interface" concept since the chains
 * only deal with INPUT (destined for machine), OUTPUT (originating from machine),
 * and FORWARD (passing through machine) which means that there's also these
 * restrictions:
 *
 *   1) you can't gather statistics for packets destined to or originating from
 *      the machine itself
 *
 *   2) since the FORWARD chain is used, the interface name doesn't matter and is
 *      ignored
 */

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/socket.h>
#include <unistd.h>
#include <net/if.h>
#include <netinet/in.h>
#include <linux/netfilter_ipv4/ip_tables.h>

#include <debug/log.h>
#include <debug/memory.h>

#include <tinysnmp/compat/inttypes.h>

#include "fw.h"

#define ARRAYSIZE(x) (sizeof (x) / sizeof ((x)[0]))

#ifndef IPT_ALIGN
#define IPT_ALIGN(s) (((s) + (__alignof__ (struct ipt_entry) - 1)) & ~(__alignof__ (struct ipt_entry) - 1))
#endif	/* #ifndef IPT_ALIGN */

typedef struct
{
   int fd;
   struct ipt_getinfo info;
   struct ipt_get_entries *entries;
} ipt_t;

typedef struct rules_r
{
   uint16_t rule;				/* rule number									*/
   struct ipt_entry **entry;	/* list of entries associated with this rule	*/
   size_t n;					/* number of entries in list					*/
   struct rules_r *next;
} rules_t;

static rules_t *cache = NULL;

static void rules_destroy ()
{
   size_t i;
   rules_t *tmp;

   while (cache != NULL)
	 {
		for (i = 0; i < cache->n; i++) mem_free (cache->entry[i]);
		mem_free (cache->entry);
		tmp = cache;
		cache = cache->next;
		mem_free (tmp);
	 }
}

static void rules_remove (uint16_t rule)
{
   rules_t *cur = cache,*prev = NULL;

   while (cur != NULL)
	 {
		if (cur->rule == rule)
		  {
			 size_t i;

			 if (prev != NULL && cur->next != NULL)
			   prev->next = cur->next;
			 else if (prev == NULL)
			   cache = cur->next;
			 else prev->next = NULL;

			 for (i = 0; i < cur->n; i++) mem_free (cur->entry[i]);
			 mem_free (cur->entry);
			 mem_free (cur);

			 return;
		  }
		else prev = cur, cur = cur->next;
	 }
}

static int rules_add (int level,uint16_t rule,const struct ipt_entry *entry)
{
   rules_t *i;
   static int init = 0;

   for (i = cache; i != NULL; i = i->next)
	 if (i->rule == rule)
	   break;

   if (i == NULL)
	 {
		int saved;

		if ((i = (rules_t *) mem_alloc (sizeof (rules_t))) == NULL)
		  {
			 oom:
			 log_printf (level,"couldn't allocate memory: %s\n",strerror (errno));
			 return (-1);
		  }

		if ((i->entry = (struct ipt_entry **) mem_alloc (sizeof (struct ipt_entry *))) == NULL)
		  {
			 saved = errno;
			 mem_free (i);
			 errno = saved;
			 goto oom;
		  }

		if ((i->entry[0] = (struct ipt_entry *) mem_alloc (entry->next_offset)) == NULL)
		  {
			 saved = errno;
			 mem_free (i->entry);
			 mem_free (i);
			 errno = saved;
			 goto oom;
		  }

		i->rule = rule;
		i->n = 1;
		memcpy (i->entry[0],entry,entry->next_offset);
		i->next = cache;
		cache = i;
	 }
   else
	 {
		void *tmp;

		if ((tmp = mem_realloc (i->entry,sizeof (struct ipt_entry *) * (i->n + 1))) == NULL)
		  goto oom;

		i->entry = (struct ipt_entry **) tmp;

		if ((i->entry[i->n] = (struct ipt_entry *) mem_alloc (entry->next_offset)) == NULL)
		  goto oom;

		memcpy (i->entry[i->n++],entry,entry->next_offset);
	 }

   if (!init)
	 {
		init = 1;
		atexit (rules_destroy);
	 }

   return (0);
}

#define ADDR(x) (x) & 0x000000ff,((x) & 0x0000ff00) >> 8,((x) & 0x00ff0000) >> 16,((x) & 0xff000000) >> 24
#define PROTO(x) ((x) == IPPROTO_TCP ? "tcp" : ((x) == IPPROTO_UDP ? "udp" : ((x) == IPPROTO_IP ? "ip" : "unknown")))
#define VERDICT(x) ((x) == -NF_ACCEPT - 1 ? "NF_ACCEPT" : ((x) == -NF_DROP - 1 ? "NF_DROP" : ((x) == -NF_QUEUE - 1 ? "NF_QUEUE" : ((x) == IPT_RETURN ? "IPT_RETURN" : "UNKNOWN"))))

#ifdef DEBUG
static void dump_target (const char *filename,int line,const char *function,int level,const struct ipt_entry_target *target)
{
   log_printf_stub (filename,line,function,level,"target name: `%s' [%u]\n",target->u.user.name,target->u.target_size);
   if (!strcmp (target->u.user.name,IPT_STANDARD_TARGET))
	 {
		int pos = *(int *) target->data;

		if (pos < 0)
		  log_printf_stub (filename,line,function,level,"verdict=%s\n",VERDICT (pos));
		else
		  log_printf_stub (filename,line,function,level,"verdict=%u\n",pos);
	 }
   else if (!strcmp (target->u.user.name,IPT_ERROR_TARGET))
	 log_printf_stub (filename,line,function,level,"error=`%s'\n",target->data);
}

static void print_ports (const char *filename,int line,const char *function,int level,const char *name,uint16_t min,uint16_t max,int invert)
{
   const char *inv = invert ? "!" : "";

   if (min || max != 0xffff || invert)
	 {
		log_puts_stub (filename,line,function,level,name);

		if (min == max)
		  log_printf_stub (filename,line,function,level,":%s%u",inv,min);
		else
		  log_printf_stub (filename,line,function,level,"s:%s%u:%u",inv,min,max);

		log_putc_stub (filename,line,function,level,' ');
	 }
}

static void print_option (const char *filename,int line,const char *function,int level,uint8_t option,int invert)
{
   if (option || invert)
	 log_printf_stub (filename,line,function,level,"option=%s%u ",invert ? "!" : "",option);
}

static void print_flags (const char *filename,int line,const char *function,int level,uint8_t mask,uint8_t cmp,int invert)
{
   if (mask || invert)
	 log_printf_stub (filename,line,function,level,"flags:%s%.2x/%.2x ",invert ? "!" : "",mask,cmp);
}

static int dump_match_tcp (const char *filename,int line,const char *function,int level,const struct ipt_tcp *tcp)
{
   log_printf_stub (filename,line,function,level,"tcp ");
   print_ports (filename,line,function,level,"spt",tcp->spts[0],tcp->spts[1],tcp->invflags & IPT_TCP_INV_SRCPT);
   print_ports (filename,line,function,level,"dpt",tcp->dpts[0],tcp->dpts[1],tcp->invflags & IPT_TCP_INV_DSTPT);
   print_option (filename,line,function,level,tcp->option,tcp->invflags & IPT_TCP_INV_OPTION);
   print_flags (filename,line,function,level,tcp->flg_mask,tcp->flg_cmp,tcp->invflags & IPT_TCP_INV_FLAGS);
   if (tcp->invflags & ~IPT_TCP_INV_MASK)
	 log_printf_stub (filename,line,function,level,"unknown invflags %.2X ",tcp->invflags & ~IPT_TCP_INV_MASK);
   log_putc_stub (filename,line,function,level,'\n');
   return (0);
}

static int dump_match_udp (const char *filename,int line,const char *function,int level,const struct ipt_udp *udp)
{
   log_printf_stub (filename,line,function,level,"udp ");
   print_ports (filename,line,function,level,"spt",udp->spts[0],udp->spts[1],udp->invflags & IPT_UDP_INV_SRCPT);
   print_ports (filename,line,function,level,"dpt",udp->dpts[0],udp->dpts[1],udp->invflags & IPT_UDP_INV_DSTPT);
   if (udp->invflags & ~IPT_UDP_INV_MASK)
	 log_printf_stub (filename,line,function,level,"unknown invflags %.2X ",udp->invflags & ~IPT_UDP_INV_MASK);
   log_putc_stub (filename,line,function,level,'\n');
   return (0);
}

static int dump_match (const struct ipt_entry_match *match,const char *filename,int line,const char *function,int level)
{
   log_printf_stub (filename,line,function,level,"match name: `%s' [%u]\n",match->u.user.name,match->u.match_size);

   if (!strcmp (match->u.user.name,"tcp"))
	 return (dump_match_tcp (filename,line,function,level,(const struct ipt_tcp *) match->data));

   if (!strcmp (match->u.user.name,"udp"))
	 return (dump_match_udp (filename,line,function,level,(const struct ipt_udp *) match->data));

   return (0);
}

#define dump_entry(level,entry) dump_entry_stub(entry,__FILE__,__LINE__,__FUNCTION__,level)
static int dump_entry_stub (const struct ipt_entry *entry,const char *filename,int line,const char *function,int level)
{
   int i,flags;
   log_printf_stub (filename,line,function,level,
			   "ENTRY\n"
			   "=========================================\n"
			   "src ip: %u.%u.%u.%u/%u.%u.%u.%u\n"
			   "dst ip: %u.%u.%u.%u/%u.%u.%u.%u\n"
			   "iface: `%s'/",
			   ADDR (entry->ip.src.s_addr),ADDR (entry->ip.smsk.s_addr),
			   ADDR (entry->ip.dst.s_addr),ADDR (entry->ip.dmsk.s_addr),
			   entry->ip.iniface);

   for (i = 0; i < IFNAMSIZ; i++) log_putc_stub (filename,line,function,level,entry->ip.iniface_mask[i] ? 'X' : '.');
   log_printf_stub (filename,line,function,level," to `%s'/",entry->ip.outiface);
   for (i = 0; i < IFNAMSIZ; i++) log_putc_stub (filename,line,function,level,entry->ip.outiface_mask[i] ? 'X' : '.');

   log_printf_stub (filename,line,function,level,
			   "\n"
			   "proto: %u (%s)\n"
			   "flags: %.2X\n"
			   "invflags: %.2X\n"
			   "counters: %llu packets, %llu bytes\n"
			   "nfcache: ",
			   entry->ip.proto,PROTO (entry->ip.proto),
			   entry->ip.flags,
			   entry->ip.invflags,
			   entry->counters.pcnt,entry->counters.bcnt);

   flags = entry->nfcache;

#define FLAGS(x)												\
   if (flags & NFC_##x)											\
	 {															\
		log_puts_stub (filename,line,function,level,#x " ");	\
		flags &= ~NFC_##x;										\
	 }

   FLAGS (ALTERED);
   FLAGS (UNKNOWN);
   FLAGS (IP_SRC);
   FLAGS (IP_DST);
   FLAGS (IP_IF_IN);
   FLAGS (IP_IF_OUT);
   FLAGS (IP_TOS);
   FLAGS (IP_PROTO);
   FLAGS (IP_OPTIONS);
   FLAGS (IP_TCPFLAGS);
   FLAGS (IP_SRC_PT);
   FLAGS (IP_DST_PT);
   FLAGS (IP_PROTO_UNKNOWN);

   if (flags) log_printf_stub (filename,line,function,level,"%.8X",flags);

   log_putc_stub (filename,line,function,level,'\n');

   IPT_MATCH_ITERATE (entry,dump_match,filename,line,function,level);

   dump_target (filename,line,function,level,(struct ipt_entry_target *) ((uint8_t *) entry + entry->target_offset));

   log_putc_stub (filename,line,function,level,'\n');

   return (0);
}

#define dump_replace(level,replace) dump_replace_stub(__FILE__,__LINE__,__FUNCTION__,level,replace)
static void dump_replace_stub (const char *filename,int line,const char *function,int level,const struct ipt_replace *replace)
{
   unsigned int i;

   log_printf_stub (filename,line,function,level,
					"replace\n"
					"===================================\n"
					"name:         `%s'\n"
					"valid_hooks:  0x%.8x\n"
					"num_entries:  %u\n"
					"size:         %u\n"
					"hook_entry:  ",
					replace->name,
					replace->valid_hooks,
					replace->num_entries,
					replace->size);

   for (i = 0; i < NF_IP_NUMHOOKS; i++)
	 log_printf_stub (filename,line,function,level," %u",replace->hook_entry[i]);

   log_puts_stub (filename,line,function,level,"\nunderflow:   ");

   for (i = 0; i < NF_IP_NUMHOOKS; i++)
	 log_printf_stub (filename,line,function,level," %u",replace->underflow[i]);

   log_printf_stub (filename,line,function,level,
					"\nnum_counters: %u\n"
					"counters:\n",
					replace->num_counters);

   for (i = 0; i < replace->num_counters; i++)
	 log_printf_stub (filename,line,function,level,
					  "    pcnt: %-10llu bcnt %llu\n",
					  replace->counters[i].pcnt,
					  replace->counters[i].bcnt);

   log_puts_stub (filename,line,function,level,"entries:\n");

   IPT_ENTRY_ITERATE (replace->entries,replace->size,dump_entry_stub,filename,line,function,level);
}
#endif	/* #ifdef DEBUG */

static int ipt_open (int level,int chain,ipt_t *ipt)
{
   socklen_t len;

   ipt->entries = NULL;

   if ((ipt->fd = socket (PF_INET,SOCK_RAW,IPPROTO_RAW)) < 0)
	 {
		log_printf (level,"socket: %s\n",strerror (errno));
		fail:
		if (ipt->fd > 0) close (ipt->fd);
		ipt->fd = -1;
		return (-1);
	 }

   strcpy (ipt->info.name,"filter");
   len = sizeof (struct ipt_getinfo);

   if (getsockopt (ipt->fd,IPPROTO_IP,IPT_SO_GET_INFO,&ipt->info,&len) < 0)
	 {
		log_printf (level,"getsockopt IPT_SO_GET_INFO: %s\n",strerror (errno));
		goto fail;
	 }

   if (!(ipt->info.valid_hooks & (1 << chain)))
	 {
		switch (chain)
		  {
#define CHAIN(x) case x: log_puts (level,#x); break
			 CHAIN (NF_IP_PRE_ROUTING);
			 CHAIN (NF_IP_LOCAL_IN);
			 CHAIN (NF_IP_FORWARD);
			 CHAIN (NF_IP_LOCAL_OUT);
			 CHAIN (NF_IP_POST_ROUTING);
#undef CHAIN
		   default:
			 log_printf (level,"hook %u",chain);
		  }

		log_puts (level," is not valid!\n");
		goto fail;
	 }

   len = ipt->info.size + sizeof (struct ipt_get_entries);

   if ((ipt->entries = (struct ipt_get_entries *) mem_alloc (len)) == NULL)
	 {
		log_printf (level,"unable to allocate memory: %s\n",strerror (errno));
		goto fail;
	 }

   strcpy (ipt->entries->name,ipt->info.name);
   ipt->entries->size = ipt->info.size;

   if (getsockopt (ipt->fd,IPPROTO_IP,IPT_SO_GET_ENTRIES,ipt->entries,&len) < 0)
	 {
		log_printf (level,"getsockopt IPT_SO_GET_ENTRIES: %s\n",strerror (errno));
		mem_free (ipt->entries);
		goto fail;
	 }

   return (0);
}

static void ipt_close (ipt_t *ipt)
{
   if (ipt->fd > 0)
	 {
		close (ipt->fd);
		ipt->fd = -1;
	 }

   if (ipt->entries != NULL)
	 {
		mem_free (ipt->entries);
		ipt->entries = NULL;
	 }
}

static int ipt_match (unsigned int offset,const struct ipt_entry *haystack,const struct ipt_entry *needle)
{
   unsigned int i;
   struct ipt_standard_target *ht,*nt;
   struct ipt_entry_match *hm,*nm;

   ht = (struct ipt_standard_target *) ((uint8_t *) haystack + haystack->target_offset);
   nt = (struct ipt_standard_target *) ((uint8_t *) needle + needle->target_offset);

   if (memcmp (&haystack->ip,&needle->ip,sizeof (struct ipt_ip)) ||
	   haystack->nfcache != needle->nfcache ||
	   haystack->target_offset != needle->target_offset ||
	   haystack->next_offset != needle->next_offset ||
	   memcmp (&haystack->counters,&needle->counters,sizeof (struct ipt_counters)) ||
	   strcmp (ht->target.u.user.name,IPT_STANDARD_TARGET) ||
	   strcmp (nt->target.u.user.name,IPT_STANDARD_TARGET) ||
	   ht->verdict < offset || ht->verdict - offset != haystack->next_offset)
	 return (0);

   for (i = sizeof (struct ipt_entry); i < haystack->target_offset; i += hm->u.match_size)
	 {
		hm = (struct ipt_entry_match *) ((uint8_t *) haystack + i);
		nm = (struct ipt_entry_match *) ((uint8_t *) needle + i);

		if (hm->u.match_size != nm->u.match_size ||
			hm->u.user.match_size != nm->u.user.match_size ||
			strcmp (hm->u.user.name,nm->u.user.name))
		  return (0);

		if (!strcmp (hm->u.user.name,"udp"))
		  {
			 if (memcmp (hm->data,nm->data,sizeof (struct ipt_udp)))
			   return (0);
		  }
		else if (!strcmp (hm->u.user.name,"tcp"))
		  {
			 if (memcmp (hm->data,nm->data,sizeof (struct ipt_tcp)))
			   return (0);
		  }
	 }

   return (1);
}

static void set_verdict (struct ipt_entry *entries,unsigned int size,unsigned int offset,unsigned int delta)
{
   struct ipt_standard_target *target;
   struct ipt_entry *tmp;
   unsigned int i,cur;

   for (i = 0; i < size; i += tmp->next_offset)
	 {
		tmp = (struct ipt_entry *) ((uint8_t *) entries + i);
		target = (struct ipt_standard_target *) ((uint8_t *) tmp + tmp->target_offset);
		cur = (uint8_t *) tmp - (uint8_t *) entries;

		if (!strcmp (target->target.u.user.name,IPT_STANDARD_TARGET) &&
			target->verdict > (int) offset &&
			(cur != offset || target->verdict != cur + tmp->next_offset))
		  target->verdict += delta;
	 }
}

static int ipt_delete (int level,unsigned int chain,struct ipt_entry *entry)
{
   struct ipt_replace *replace;
   struct ipt_entry *tmp = NULL;
   ipt_t ipt;
   unsigned int i,offset;
   socklen_t len;

#ifdef DEBUG
   dump_entry (LOG_DEBUG,entry);
#endif	/* #ifdef DEBUG */

   if (ipt_open (level,chain,&ipt) < 0)
	 return (-1);

   for (offset = 0; offset < ipt.info.size; offset += tmp->next_offset)
	 {
		tmp = (struct ipt_entry *) ((uint8_t *) ipt.entries->entrytable + offset);
		if (ipt_match (offset,tmp,entry))
		  break;
	 }

   if (tmp == NULL || offset == ipt.info.size)
	 {
		log_printf (level,"entry not found!\n");
		ipt_close (&ipt);
		return (-1);
	 }

   len = sizeof (struct ipt_replace) + ipt.info.size - tmp->next_offset;

   if ((replace = (struct ipt_replace *) mem_alloc (len + sizeof (struct ipt_counters) * ipt.info.num_entries)) == NULL)
	 {
		log_printf (level,"unable to allocate memory: %s\n",strerror (errno));
		ipt_close (&ipt);
		return (-1);
	 }

   strcpy (replace->name,ipt.info.name);
   replace->valid_hooks = ipt.info.valid_hooks;
   replace->num_entries = ipt.info.num_entries - 1;
   replace->size = ipt.info.size - tmp->next_offset;
   replace->num_counters = ipt.info.num_entries;
   replace->counters = (struct ipt_counters *) ((uint8_t *) replace + len);

   for (i = 0; i < NF_IP_NUMHOOKS; i++)
	 {
		replace->hook_entry[i] = ipt.info.hook_entry[i];
		replace->underflow[i] = ipt.info.underflow[i];

		if (replace->hook_entry[i] >= offset + tmp->next_offset)
		  replace->hook_entry[i] -= tmp->next_offset;

		if (replace->underflow[i] >= offset + tmp->next_offset)
		  replace->underflow[i] -= tmp->next_offset;
	 }

   memcpy (replace->entries,ipt.entries->entrytable,offset);
   memcpy ((uint8_t *) replace->entries + offset,
		   (uint8_t *) ipt.entries->entrytable + offset + tmp->next_offset,
		   replace->size - offset);

   set_verdict (replace->entries,replace->size,offset,-entry->next_offset);

#ifdef DEBUG
   dump_replace (LOG_NOISY,replace);
#endif	/* #ifdef DEBUG */

   if (setsockopt (ipt.fd,IPPROTO_IP,IPT_SO_SET_REPLACE,replace,len) < 0)
	 {
		log_printf (level,"setsockopt IPT_SO_SET_REPLACE: %s\n",strerror (errno));
		ipt_close (&ipt);
		mem_free (replace);
		return (-1);
	 }

   mem_free (replace);
   ipt_close (&ipt);

   return (0);
}

static int ipt_add (int level,unsigned int chain,struct ipt_entry *entry)
{
   struct ipt_standard_target *target = (struct ipt_standard_target *) ((uint8_t *) entry + entry->target_offset);
   struct ipt_replace *replace;
   socklen_t len;
   unsigned int i;
   ipt_t ipt;

#ifdef DEBUG
   dump_entry (LOG_DEBUG,entry);
#endif	/* #ifdef DEBUG */

   if (ipt_open (level,chain,&ipt) < 0)
	 return (-1);

   target->verdict = ipt.info.hook_entry[chain] + entry->next_offset;

   len = sizeof (struct ipt_replace) + ipt.info.size + entry->next_offset;

   if ((replace = (struct ipt_replace *) mem_alloc (len + sizeof (struct ipt_counters) * ipt.info.num_entries)) == NULL)
	 {
		log_printf (level,"unable to allocate memory: %s\n",strerror (errno));
		ipt_close (&ipt);
		return (-1);
	 }

   strcpy (replace->name,ipt.info.name);
   replace->valid_hooks = ipt.info.valid_hooks;
   replace->num_entries = ipt.info.num_entries + 1;
   replace->size = ipt.info.size + entry->next_offset;
   replace->num_counters = ipt.info.num_entries;
   replace->counters = (struct ipt_counters *) ((uint8_t *) replace + len);

   for (i = 0; i < NF_IP_NUMHOOKS; i++)
	 {
		replace->hook_entry[i] = ipt.info.hook_entry[i];
		replace->underflow[i] = ipt.info.underflow[i];

		if (replace->hook_entry[i] > ipt.info.hook_entry[chain])
		  replace->hook_entry[i] += entry->next_offset;

		if (replace->underflow[i] >= ipt.info.hook_entry[chain])
		  replace->underflow[i] += entry->next_offset;
	 }

   memcpy (replace->entries,ipt.entries->entrytable,ipt.info.hook_entry[chain]);
   memcpy ((uint8_t *) replace->entries + ipt.info.hook_entry[chain],entry,entry->next_offset);
   memcpy ((uint8_t *) replace->entries + ipt.info.hook_entry[chain] + entry->next_offset,
		   (uint8_t *) ipt.entries->entrytable + ipt.info.hook_entry[chain],
		   ipt.info.size - ipt.info.hook_entry[chain]);

   set_verdict (replace->entries,replace->size,ipt.info.hook_entry[chain],entry->next_offset);

#ifdef DEBUG
   dump_replace (LOG_NOISY,replace);
#endif	/* #ifdef DEBUG */

   if (setsockopt (ipt.fd,IPPROTO_IP,IPT_SO_SET_REPLACE,replace,len) < 0)
	 {
		log_printf (level,"setsockopt IPT_SO_SET_REPLACE: %s\n",strerror (errno));
		ipt_close (&ipt);
		mem_free (replace);
		return (-1);
	 }

   mem_free (replace);
   ipt_close (&ipt);

   return (0);
}

/*
 * Add rules to count traffic that matches the specified criteria. Returns 0
 * if successful, -1 otherwise. The function automatically logs errors.
 *
 *        level     Log level
 *        n         Rule number
 *        proto     IP protocol (IPPROTO_IP, IPPROTO_TCP, or IPPROTO_UDP)
 *        net       Source/Destination network to match
 *        port      Port number (if 0, any port matches)
 *        iface     Name of interface on which to match packets (unused)
 *        src       If 0, match destination, else match source
 */
int fw_add_counter (int level,uint16_t n,int proto,const struct network *net,uint16_t port,const char *iface,int src)
{
   static const struct
	 {
		unsigned int chain;
		unsigned int flags;
	 } mask[] =
	 {	/* WARNING: don't change this */
		{ NF_IP_FORWARD, NFC_IP_SRC_PT },
		{ NF_IP_FORWARD, NFC_IP_DST_PT }
	 };
   struct ipt_entry *entry;
   struct ipt_standard_target *target;
   struct ipt_entry_match *match = NULL;
   struct ipt_tcp *tcp = NULL;
   struct ipt_udp *udp = NULL;
   size_t tmp,match_size = 0,target_size = IPT_ALIGN (sizeof (struct ipt_standard_target));
   int i;

   if (proto == IPPROTO_UDP)
	 match_size = IPT_ALIGN (sizeof (struct ipt_entry_match)) + IPT_ALIGN (sizeof (struct ipt_udp));
   else if (proto == IPPROTO_TCP)
	 match_size = IPT_ALIGN (sizeof (struct ipt_entry_match)) + IPT_ALIGN (sizeof (struct ipt_tcp));

   if ((entry = (struct ipt_entry *) mem_alloc (sizeof (struct ipt_entry) + match_size + target_size)) == NULL)
	 {
		log_printf (level,"couldn't allocate memory: %s\n",strerror (errno));
		return (-1);
	 }

   memset (entry,0L,sizeof (struct ipt_entry) + match_size + target_size);

   entry->target_offset = sizeof (struct ipt_entry);

   if (proto == IPPROTO_UDP)
	 {
		match = (struct ipt_entry_match *) ((uint8_t *) entry + entry->target_offset);
		udp = (struct ipt_udp *) match->data;
		strcpy (match->u.user.name,"udp");
	 }
   else if (proto == IPPROTO_TCP)
	 {
		match = (struct ipt_entry_match *) ((uint8_t *) entry + entry->target_offset);
		tcp = (struct ipt_tcp *) match->data;
		strcpy (match->u.user.name,"tcp");
	 }

   if (match != NULL)
	 match->u.match_size = match->u.user.match_size = match->u.kernel.match_size = match_size;

   entry->target_offset += match_size;
   entry->next_offset = entry->target_offset + target_size;

   target = (struct ipt_standard_target *) ((uint8_t *) entry + entry->target_offset);
   target->target.u.target_size = target->target.u.user.target_size = target->target.u.kernel.target_size = target_size;
   memset (target->target.u.user.name,0L,IPT_FUNCTION_MAXNAMELEN);
   strcpy (target->target.u.user.name,IPT_STANDARD_TARGET);

   entry->ip.proto = proto;

   if (proto == IPPROTO_TCP || proto == IPPROTO_UDP)
	 entry->nfcache |= NFC_IP_PROTO;

   if (src)
	 {
		entry->ip.src.s_addr = net->address;
		entry->ip.smsk.s_addr = net->netmask;
		entry->nfcache |= NFC_IP_SRC;
	 }
   else
	 {
		entry->ip.dst.s_addr = net->address;
		entry->ip.dmsk.s_addr = net->netmask;
		entry->nfcache |= NFC_IP_DST;
	 }

   tmp = ARRAYSIZE (mask);
   if (proto == IPPROTO_IP) tmp /= 2;

   for (i = 0; i < tmp; i++)
	 {
		memset (entry->ip.iniface,0L,IFNAMSIZ);
		memset (entry->ip.outiface,0L,IFNAMSIZ);
		memset (entry->ip.iniface_mask,0L,IFNAMSIZ);
		memset (entry->ip.outiface_mask,0L,IFNAMSIZ);
		entry->nfcache &= ~(NFC_IP_IF_IN | NFC_IP_IF_OUT | NFC_IP_SRC_PT | NFC_IP_DST_PT);

		if (mask[i].chain == NF_IP_LOCAL_IN)
		  {
			 strcpy (entry->ip.iniface,iface);
			 memset (entry->ip.iniface_mask,0xff,strlen (iface) + 1);
			 entry->nfcache |= NFC_IP_IF_IN;
		  }

		if (mask[i].chain == NF_IP_LOCAL_OUT)
		  {
			 strcpy (entry->ip.outiface,iface);
			 memset (entry->ip.outiface_mask,0xff,strlen (iface) + 1);
			 entry->nfcache |= NFC_IP_IF_OUT;
		  }

		if (proto == IPPROTO_UDP)
		  {
			 entry->nfcache |= mask[i].flags;

			 if (mask[i].flags == NFC_IP_SRC_PT)
			   {
				  udp->spts[0] = udp->spts[1] = port;
				  udp->dpts[0] = 0, udp->dpts[1] = 0xffff;
			   }
			 else
			   {
				  udp->spts[0] = 0, udp->spts[1] = 0xffff;
				  udp->dpts[0] = udp->dpts[1] = port;
			   }
		  }
		else if (proto == IPPROTO_TCP)
		  {
			 entry->nfcache |= mask[i].flags;

			 if (mask[i].flags == NFC_IP_SRC_PT)
			   {
				  tcp->spts[0] = tcp->spts[1] = port;
				  tcp->dpts[0] = 0, tcp->dpts[1] = 0xffff;
			   }
			 else
			   {
				  tcp->spts[0] = 0, tcp->spts[1] = 0xffff;
				  tcp->dpts[0] = tcp->dpts[1] = port;
			   }
		  }

		if (rules_add (level,n,entry) < 0)
		  {
			 mem_free (entry);
			 return (-1);
		  }

		if (ipt_add (level,mask[i].chain,entry) < 0)
		  {
			 mem_free (entry);
			 rules_remove (n);
			 return (-1);
		  }
	 }

   mem_free (entry);

   return (0);
}

/*
 * Delete rules that match the specified rule number. Returns 0 if
 * successful, -1 otherwise. The function automatically logs errors.
 *
 *        level     Log level
 *        n         Rule number
 */
int fw_del_counters (int level,uint16_t n)
{
   rules_t *tmp;

   for (tmp = cache; tmp != NULL; tmp = tmp->next)
	 if (tmp->rule == n)
	   {
		  size_t i,errors = 0;

		  for (i = 0; i < tmp->n; i++)
			if (ipt_delete (level,NF_IP_FORWARD,tmp->entry[i]) < 0)
			  errors++;

		  return (errors ? -1 : 0);
	   }

   return (0);
}

/*
 * Get a snapshot of the current firewall rules. This is needed by
 * fw_get_counters(). Returns a pointer to the allocated data structure
 * of NULL if some error occurred. The function automatically logs errors.
 *
 *        level     Log level
 */
void *fw_get_rules (int level)
{
   ipt_t *ipt;

   if ((ipt = (ipt_t *) mem_alloc (sizeof (ipt_t))) == NULL)
	 {
		log_printf (level,"couldn't allocate memory: %s\n",strerror (errno));
		return (NULL);
	 }

   if (ipt_open (level,NF_IP_FORWARD,ipt) < 0)
	 {
		mem_free (ipt);
		return (NULL);
	 }

   return (ipt);
}

/*
 * Return the sum of the interface byte count counters of the rules
 * that match the specified rule number. Returns 0 if one of more
 * rules matched, or -1 if no firewall rules match the specified
 * rule number.
 *
 *        rules     List of rules allocated by fw_get_rules()
 *        n         Rule number
 *        count     Sum of matched rule byte counters
 */
int fw_get_counters (void *rules,uint16_t n,uint64_t *count)
{
   rules_t *tmp;

   for (tmp = cache; tmp != NULL; tmp = tmp->next)
	 if (tmp->rule == n)
	   {
		  ipt_t *ipt = (ipt_t *) rules;
		  size_t i;
		  unsigned int offset;
		  struct ipt_entry *entry;

		  *count = 0;

		  for (i = 0; i < tmp->n; i++)
			for (offset = 0; offset < ipt->info.size; offset += entry->next_offset)
			  {
				 entry = (struct ipt_entry *) ((uint8_t *) ipt->entries->entrytable + offset);
				 if (ipt_match (offset,entry,tmp->entry[i]))
				   {
					  *count += entry->counters.bcnt;
					  break;
				   }
			  }

		  return (0);
	   }

   return (-1);
}

/*
 * Free memory allocated by the fw_get_rules() function.
 *
 *        rules     List of rules allocated by fw_get_rules()
 */
void fw_destroy_rules (void *rules)
{
   if (rules != NULL)
	 {
		ipt_close ((ipt_t *) rules);
		mem_free (rules);
	 }
}

#if 0
int main ()
{
   struct network net;
   void *tmp;
   uint64_t count;

   mem_open (NULL);
   log_open (NULL,LOG_NOISY,LOG_HAVE_COLORS);
   atexit (mem_close);
   atexit (log_close);

   net.address = 0x42081c33;
   net.netmask = 0xffffff00;

   if (fw_add_counter (LOG_ERROR,0,IPPROTO_IP,&net,0,"eth0",0) < 0)
	 exit (EXIT_FAILURE);

   if (fw_add_counter (LOG_ERROR,1,IPPROTO_UDP,&net,25,"eth0",0) < 0)
	 exit (EXIT_FAILURE);

   system ("iptables -L -n -v");

   if ((tmp = fw_get_rules (LOG_ERROR)) == NULL)
	 {
		fw_del_counters (LOG_ERROR,0);
		exit (EXIT_FAILURE);
	 }

   if (fw_get_counters (tmp,0,&count) < 0)
	 {
		log_printf (LOG_ERROR,"rule #0 not found!\n");
		fw_del_counters (LOG_ERROR,0);
		exit (EXIT_FAILURE);
	 }

   fw_destroy_rules (tmp);

   if (fw_del_counters (LOG_ERROR,0) < 0 || fw_del_counters (LOG_ERROR,1) < 0)
	 exit (EXIT_FAILURE);

   log_printf (LOG_DEBUG,"rule #0 == %llu\n",count);

   system ("iptables -L -n -v");
   system ("iptables -F");

   exit (EXIT_SUCCESS);
}
#endif

