
/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <time.h>

#include <debug/log.h>
#include <tinysnmp/compat/inttypes.h>
#include <abz/contigmask.h>

#include "config.h"
#include "fw.h"
#include "rules.h"

#define ADDR(x) (x) & 0x000000ff,((x) & 0x0000ff00) >> 8,((x) & 0x00ff0000) >> 16,((x) & 0xff000000) >> 24
#define ARRAYSIZE(x) (sizeof (x) / sizeof ((x)[0]))
#define PROTO(x) ((x) == IPPROTO_IP ? "ip" : ((x) == IPPROTO_TCP ? "tcp" : "udp"))

typedef struct
{
   int type;
   int proto;
   int port;
} proto_mapping_t;

/* protocol mapping */
static proto_mapping_t protocols[] =
{
   { ftp, IPPROTO_TCP, 20 },
   { ftp, IPPROTO_TCP, 21 },
   { ssh, IPPROTO_TCP, 22 },
   { ssh, IPPROTO_UDP, 22 },
   { telnet, IPPROTO_TCP, 23 },
   { smtp, IPPROTO_TCP, 25 },
   { dns, IPPROTO_TCP, 53 },
   { dns, IPPROTO_UDP, 53 },
   { tftp, IPPROTO_UDP, 69 },
   { http, IPPROTO_TCP, 80 },
   { http, IPPROTO_UDP, 80 },
   { pop3, IPPROTO_TCP, 110 },
   { pop3, IPPROTO_UDP, 110 },
   { imap, IPPROTO_TCP, 143 },
   { imap, IPPROTO_UDP, 143 },
   { imap, IPPROTO_TCP, 220 },
   { imap, IPPROTO_UDP, 220 },
   { snmp, IPPROTO_TCP, 161 },
   { snmp, IPPROTO_UDP, 161 },
   { snmp, IPPROTO_TCP, 162 },
   { snmp, IPPROTO_UDP, 162 },
   { irc, IPPROTO_TCP, 194 },
   { irc, IPPROTO_UDP, 194 },
   { irc, IPPROTO_TCP, 6667 },
   { irc, IPPROTO_UDP, 6667 },
   { total, IPPROTO_IP, 0 }
};

/*
 * Remove all the firewall rules that was added for each client in
 * the traffic structure. Warnings are logged automatically.
 */
void rules_remove (traffic_t *traffic)
{
   uint32_t client,protocol;

   for (client = 0; client < traffic->n; client++)
	 for (protocol = total; protocol < traffic->clients[client].n; protocol++)
	   {
		  if (traffic->clients[client].recv[protocol].rule >= 0)
			{
			   if (fw_del_counters (LOG_WARNING,traffic->clients[client].recv[protocol].rule) < 0)
				 log_printf (LOG_WARNING,"Failed to remove accounting rule %" PRId32 "\n",traffic->clients[client].recv[protocol].rule);

			   traffic->clients[client].recv[protocol].rule = -1;
			}

		  if (traffic->clients[client].xmit[protocol].rule >= 0)
			{
			   if (fw_del_counters (LOG_WARNING,traffic->clients[client].xmit[protocol].rule) < 0)
				 log_printf (LOG_WARNING,"Failed to remove accounting rule %" PRId32 "\n",traffic->clients[client].xmit[protocol].rule);

			   traffic->clients[client].xmit[protocol].rule = -1;
			}
	   }
}

/*
 * Add accounting rules for each client in the specified
 * traffic structure. Returns 0 if successful, -1 if some
 * error occurred. Errors are logged automatically.
 */
int rules_add (traffic_t *traffic)
{
   uint32_t client,protocol,port,network;
   uint16_t rule = traffic->rulestart;

   /* mark all counters as invalid */
   for (client = 0; client < traffic->n; client++)
	 for (protocol = unknown; protocol < traffic->clients[client].n; protocol++)
	   {
		  traffic->clients[client].recv[protocol].rule = traffic->clients[client].xmit[protocol].rule = -1;
		  traffic->clients[client].recv[protocol].valid = traffic->clients[client].xmit[protocol].valid = 0;
	   }

   /* now try to update them */
   for (client = 0; client < traffic->n; client++)
	 for (protocol = total; protocol < traffic->clients[client].n; protocol++)
	   {
		  /* add counters for traffic originating at the client */

		  for (port = 0; port < ARRAYSIZE (protocols); port++)
			if (protocols[port].type == traffic->clients[client].xmit[protocol].proto)
			  for (network = 0; network < traffic->clients[client].info.n; network++)
				if (fw_add_counter (LOG_ERROR,
									  rule,
									  protocols[port].proto,
									  traffic->clients[client].info.networks + network,
									  protocols[port].port,
									  traffic->interface,
									  MATCH_OUTGOING) < 0)
				  {
					 uint8_t mask;

					 log_printf (LOG_ERROR,
								 "Failed to add the following accounting rule:\n"
								 "    fw add %" PRIu16 " count %s from %" PRIu32 ".%" PRIu32 ".%" PRIu32 ".%" PRIu32 "/",
								 rule,
								 PROTO (protocols[port].proto),
								 ADDR (traffic->clients[client].info.networks[network].address));

					 if ((mask = contigmask (traffic->clients[client].info.networks[network].netmask)) < 0xff)
					   log_printf (LOG_ERROR,
								   "%" PRIu8,
								   mask);
					 else
					   log_printf (LOG_ERROR,
								   "%" PRIu32 ".%" PRIu32 ".%" PRIu32 ".%" PRIu32,
								   ADDR (traffic->clients[client].info.networks[network].netmask));

					 if (protocols[port].port) log_printf (LOG_ERROR," %d",protocols[port].port);

					 log_printf (LOG_ERROR," to any via %s\n",traffic->interface);

					 rules_remove (traffic);

					 return (-1);
				  }

		  traffic->clients[client].xmit[protocol].rule = rule++;

		  /* add counters for traffic destined for the client */

		  for (port = 0; port < ARRAYSIZE (protocols); port++)
			if (protocols[port].type == traffic->clients[client].recv[protocol].proto)
			  for (network = 0; network < traffic->clients[client].info.n; network++)
				if (fw_add_counter (LOG_ERROR,
									  rule,
									  protocols[port].proto,
									  traffic->clients[client].info.networks + network,
									  protocols[port].port,
									  traffic->interface,
									  MATCH_INCOMING) < 0)
				  {
					 uint8_t mask;

					 log_printf (LOG_ERROR,
								 "Failed to add the following accounting rule:\n"
								 "    fw add %" PRIu16 " count %s from any to %" PRIu32 ".%" PRIu32 ".%" PRIu32 ".%" PRIu32 "/",
								 rule,
								 PROTO (protocols[port].proto),
								 ADDR (traffic->clients[client].info.networks[network].address));

					 if ((mask = contigmask (traffic->clients[client].info.networks[network].netmask)) < 0xff)
					   log_printf (LOG_ERROR,
								   "%" PRIu8,
								   mask);
					 else
					   log_printf (LOG_ERROR,
								   "%" PRIu32 ".%" PRIu32 ".%" PRIu32 ".%" PRIu32,
								   ADDR (traffic->clients[client].info.networks[network].netmask));

					 if (protocols[port].port) log_printf (LOG_ERROR," %d",protocols[port].port);

					 log_printf (LOG_ERROR," via %s\n",traffic->interface);

					 rules_remove (traffic);

					 return (-1);
				}

		  traffic->clients[client].recv[protocol].rule = rule++;
	   }

   traffic->uptime = time (NULL);

   return (0);
}

/*
 * Update traffic counters for each client in the specified
 * traffic structure. For each counter that couldn't be updated,
 * the relevant valid flag will not be set. Warnings are logged
 * automatically.
 */
void rules_update (traffic_t *traffic)
{
   void *rules;
   uint32_t client,protocol;

   /* mark all counters as invalid */
   for (client = 0; client < traffic->n; client++)
	 for (protocol = unknown; protocol < traffic->clients[client].n; protocol++)
	   traffic->clients[client].recv[protocol].valid = traffic->clients[client].xmit[protocol].valid = 0;

   /* now to try to update some/all of them */
   if ((rules = fw_get_rules (LOG_WARNING)) != NULL)
	 {
		int recv_valid,xmit_valid;
		uint64_t recv_counter,xmit_counter;

		for (client = 0; client < traffic->n; client++)
		  {
			 recv_valid = xmit_valid = 1;
			 recv_counter = xmit_counter = 0;

			 for (protocol = total; protocol < traffic->clients[client].n; protocol++)
			   {
				  if (traffic->clients[client].recv[protocol].rule >= 0 &&
					  !fw_get_counters (rules,traffic->clients[client].recv[protocol].rule,&traffic->clients[client].recv[protocol].counter))
					{
					   traffic->clients[client].recv[protocol].valid = 1;
					   if (protocol > total) recv_counter += traffic->clients[client].recv[protocol].counter;
					}
				  else
					recv_valid = 0;

				  if (traffic->clients[client].xmit[protocol].rule >= 0 &&
					  !fw_get_counters (rules,traffic->clients[client].xmit[protocol].rule,&traffic->clients[client].xmit[protocol].counter))
					{
					   traffic->clients[client].xmit[protocol].valid = 1;
					   if (protocol > total) xmit_counter += traffic->clients[client].xmit[protocol].counter;
					}
				  else
					xmit_valid = 0;
			   }

			 if (recv_valid)
			   {
				  traffic->clients[client].recv[unknown].counter = traffic->clients[client].recv[total].counter - recv_counter;
				  traffic->clients[client].recv[unknown].valid = 1;
			   }

			 if (xmit_valid)
			   {
				  traffic->clients[client].xmit[unknown].counter = traffic->clients[client].xmit[total].counter - xmit_counter;
				  traffic->clients[client].xmit[unknown].valid = 1;
			   }
		  }

		fw_destroy_rules (rules);
	 }
}

