
/*
 * Copyright (c) Abraham vd Merwe <abz@blio.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	  notice, this list of conditions and the following disclaimer in the
 *	  documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of other contributors
 *	  may be used to endorse or promote products derived from this software
 *	  without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <errno.h>
#include <fcntl.h>
#include <net/if_arp.h>
#include <stdio.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/utsname.h>
#include <unistd.h>
#include <string.h>

#include <tinysnmp/tinysnmp.h>
#include <tinysnmp/agent/mod.h>
#include <tinysnmp/agent/odb.h>
#include <tinysnmp/compat/inttypes.h>

#include <ber/ber.h>

#include <debug/log.h>
#include <debug/memory.h>

#include "proc.h"
#include "config.h"

#define NB_OF(x) (sizeof (x) / sizeof ((x)[0]))

#define IFENTRY_IFINDEX							1
#define IFENTRY_IFDESCR							2
#define IFENTRY_IFTYPE							3
#define IFENTRY_IFMTU							4
#define IFENTRY_IFSPEED							5
#define IFENTRY_IFPHYSADDRESS					6
#define IFENTRY_IFADMINSTATUS					7
#define IFENTRY_IFOPERSTATUS					8
#define IFENTRY_IFLASTCHANGE					9
#define IFENTRY_IFINOCTETS						10
#define IFENTRY_IFINUCASTPKTS					11
#define IFENTRY_IFINNUCASTPKTS					12
#define IFENTRY_IFINDISCARDS					13
#define IFENTRY_IFINERRORS						14
#define IFENTRY_IFINUNKNOWNPROTOS				15
#define IFENTRY_IFOUTOCTETS						16
#define IFENTRY_IFOUTUCASTPKTS					17
#define IFENTRY_IFOUTNUCASTPKTS					18
#define IFENTRY_IFOUTDISCARDS					19
#define IFENTRY_IFOUTERRORS						20
#define IFENTRY_IFOUTQLEN						21
#define IFENTRY_IFSPECIFIC						22

static int set_value (snmp_value_t *result,uint8_t type,void *value)
{
   result->type = type;

   switch (type)
	 {
	  case BER_INTEGER:
		result->data.INTEGER = *((int32_t *) value);
		return (0);
	  case BER_Counter32:
		result->data.Counter32 = *((uint32_t *) value);
		return (0);
	  case BER_Gauge32:
		result->data.Gauge32 = *((uint32_t *) value);
		return (0);
	  case BER_TimeTicks:
		result->data.TimeTicks = *((uint32_t *) value);
		return (0);
	  case BER_Counter64:
		result->data.Counter64 = *((uint64_t *) value);
		return (0);
	  case BER_OID:
		result->data.OID = (uint32_t *) value;
		return (0);
	  case BER_OCTET_STRING:
		  {
			 octet_string_t *tmp = (octet_string_t *) value;
			 result->data.OCTET_STRING.len = tmp->len;
			 result->data.OCTET_STRING.buf = tmp->buf;
			 return (0);
		  }
	 }

   errno = EINVAL;
   return (-1);
}

static int add_iftable (uint32_t entry,uint32_t index,uint8_t type,void *value)
{
   const uint32_t ifTable[11] = { 10, 43, 6, 1, 2, 1, 2, 2, 1, entry, index };
   snmp_value_t tmp;
   if (set_value (&tmp,type,value) < 0) return (-1);
   return (odb_add (ifTable,&tmp));
}

static int add_ifnumber (int32_t n)
{
   const uint32_t ifNumber[9] = { 8, 43, 6, 1, 2, 1, 2, 1, 0 };
   snmp_value_t value;
   if (set_value (&value,BER_INTEGER,&n) < 0) return (-1);
   return (odb_add (ifNumber,&value));
}

#define IFADD(_entry,_type,_value,_name) do {														\
	if (add_iftable (_entry,ifIndex,_type,_value) < 0)												\
	  {																								\
		 log_printf (LOG_ERROR,																		\
					 "%s: couldn't add interfaces.ifTable.ifEntry." _name ".%" PRIu32 ": %s\n",		\
					 mod->name,ifIndex,strerror (errno));											\
		 goto fail;																					\
	  }																								\
} while (0)

static int iface_update (const struct mod_r *mod)
{
   struct ifreq ifr;
   int i,fd;
   size_t n;
   devstat_t *stats;
   int32_t ifIndex;
   int32_t ifType;
   int32_t ifMtu;
   uint32_t ifSpeed;
   char ifPhysAddress[IFHWADDRLEN * 3];
   int32_t ifAdminStatus;
   int32_t ifOperStatus;
   uint32_t ifLastChange;
   uint32_t ifInOctets;
   uint32_t ifInUcastPkts;
   uint32_t ifInNUcastPkts;
   uint32_t ifInDiscards;
   uint32_t ifInErrors;
   uint32_t ifInUnknownProtos;
   uint32_t ifOutOctets;
   uint32_t ifOutUcastPkts;
   uint32_t ifOutNUcastPkts;
   uint32_t ifOutDiscards;
   uint32_t ifOutErrors;
   uint32_t ifOutQLen;
   uint32_t ifSpecific[2];
   octet_string_t str;

   odb_remove (mod->oid);

   if ((fd = socket (PF_INET,SOCK_DGRAM,0)) < 0)
	 {
		log_printf (LOG_ERROR,"%s: socket failed: %s\n",mod->name,strerror (errno));
		return (-1);
	 }

   if ((stats = getdevstats (&n)) == NULL)
	 {
		log_printf (LOG_ERROR,"%s: getdevstats failed: %s\n",mod->name,strerror (errno));
		close (fd);
		return (-1);
	 }

   if (add_ifnumber (n) < 0)
	 {
		log_printf (LOG_ERROR,"%s: couldn't add interfaces.ifNumber.0: %s\n",mod->name,strerror (errno));
		fail:
		mem_free (stats);
		close (fd);
		odb_remove (mod->oid);
		return (-1);
	 }

   for (i = 0; i < n; i++)
	 {
		if (!(ifIndex = if_nametoindex (stats[i].dev)))
		  {
			 log_printf (LOG_ERROR,"%s: couldn't map %s to an interface index\n",mod->name,stats[i].dev);
			 goto fail;
		  }

		IFADD (IFENTRY_IFINDEX,BER_INTEGER,&ifIndex,"ifIndex");

		str.len = strlen (stats[i].dev);
		str.buf = stats[i].dev;

		IFADD (IFENTRY_IFDESCR,BER_OCTET_STRING,&str,"ifDescr");

		strcpy (ifr.ifr_name,stats[i].dev);
		if (ioctl (fd,SIOCGIFHWADDR,&ifr) < 0)
		  {
			 log_printf (LOG_ERROR,"%s: ioctl SIOCGIFHWADDR failed: %s\n",mod->name,strerror (errno));
			 goto fail;
		  }

		str.len = 0;
		str.buf = ifPhysAddress;

		switch (ifr.ifr_hwaddr.sa_family)
		  {
		   case ARPHRD_ETHER:
			 ifType = 6;
			 if (IFHWADDRLEN)
			   {
				  str.len = IFHWADDRLEN;
				  memcpy (str.buf,ifr.ifr_hwaddr.sa_data,str.len);
			   }
			 break;
		   case ARPHRD_PPP:
			 ifType = 23;
			 break;
		   case ARPHRD_HDLC:
			 ifType = 118;
			 break;
		   case ARPHRD_LOOPBACK:
			 ifType = 24;
			 break;
		   default:
			 ifType = 1;
		  }

		IFADD (IFENTRY_IFPHYSADDRESS,BER_OCTET_STRING,&str,"ifPhysAddress");

		strcpy (ifr.ifr_name,stats[i].dev);
		if (ioctl (fd,SIOCGIFMTU,&ifr) < 0)
		  {
			 log_printf (LOG_ERROR,"%s: ioctl SIOCGIFMTU failed: %s\n",mod->name,strerror (errno));
			 goto fail;
		  }
		ifMtu = (int32_t) ifr.ifr_mtu;

		IFADD (IFENTRY_IFMTU,BER_INTEGER,&ifMtu,"ifMtu");

		strcpy (ifr.ifr_name,stats[i].dev);
		if (ioctl (fd,SIOCGIFFLAGS,&ifr) < 0)
		  {
			 log_printf (LOG_ERROR,"%s: ioctl SIOCGIFFLAGS failed: %s\n",mod->name,strerror (errno));
			 goto fail;
		  }
		ifAdminStatus = ifr.ifr_flags & IFF_UP ? 1 : 2;
		ifOperStatus = ifr.ifr_flags & IFF_RUNNING ? 1 : 2;

		if (ifr.ifr_flags & IFF_PORTSEL)
		  {
			 strcpy (ifr.ifr_name,stats[i].dev);
			 if (ioctl (fd,SIOCGIFMAP,&ifr) < 0)
			   {
				  log_printf (LOG_ERROR,"%s: ioctl SIOCGIFMAP failed: %s\n",mod->name,strerror (errno));
				  goto fail;
			   }
			 if (ifr.ifr_map.port == 4 || ifr.ifr_map.port == 5)
			   ifType = 62;
			 else if (ifr.ifr_map.port == 6)
			   ifType = 69;
		  }

		IFADD (IFENTRY_IFADMINSTATUS,BER_INTEGER,&ifAdminStatus,"ifAdminStatus");
		IFADD (IFENTRY_IFOPERSTATUS,BER_INTEGER,&ifOperStatus,"ifOperStatus");
		IFADD (IFENTRY_IFTYPE,BER_INTEGER,&ifType,"ifType");

		/* Linux doesn't report the speed of the interface :P */
		ifSpeed = 0xffffffff;

		IFADD (IFENTRY_IFSPEED,BER_Gauge32,&ifSpeed,"ifSpeed");

		/* ...neither does is support interface change timestamps */
		ifLastChange = 0;

		IFADD (IFENTRY_IFLASTCHANGE,BER_TimeTicks,&ifLastChange,"ifLastChange");

		ifInOctets = stats[i].rx_bytes;
		ifInUcastPkts = stats[i].rx_packets - stats[i].multicast;
		ifInNUcastPkts = stats[i].multicast;
		ifInDiscards = stats[i].rx_dropped;
		ifInErrors = stats[i].rx_errors;

		/* Linux have no way to retrieve the unknown protocol packets received */
		ifInUnknownProtos = 0;

		IFADD (IFENTRY_IFINOCTETS,BER_Counter32,&ifInOctets,"ifInOctets");
		IFADD (IFENTRY_IFINUCASTPKTS,BER_Counter32,&ifInUcastPkts,"ifInUcastPkts");
		IFADD (IFENTRY_IFINNUCASTPKTS,BER_Counter32,&ifInNUcastPkts,"ifInNUcastPkts");
		IFADD (IFENTRY_IFINDISCARDS,BER_Counter32,&ifInDiscards,"ifInDiscards");
		IFADD (IFENTRY_IFINERRORS,BER_Counter32,&ifInErrors,"ifInErrors");
		IFADD (IFENTRY_IFINUNKNOWNPROTOS,BER_Counter32,&ifInUnknownProtos,"ifInUnknownProtos");

		ifOutOctets = stats[i].tx_bytes;
		ifOutUcastPkts = stats[i].tx_packets;

		/* Linux is too brain damaged to let us know about transmitted multicast packets :P */
		ifOutNUcastPkts = 0;

		ifOutDiscards = stats[i].tx_dropped;
		ifOutErrors = stats[i].tx_errors;

		IFADD (IFENTRY_IFOUTOCTETS,BER_Counter32,&ifOutOctets,"ifOutOctets");
		IFADD (IFENTRY_IFOUTUCASTPKTS,BER_Counter32,&ifOutUcastPkts,"ifOutUcastPkts");
		IFADD (IFENTRY_IFOUTNUCASTPKTS,BER_Counter32,&ifOutNUcastPkts,"ifOutNUcastPkts");
		IFADD (IFENTRY_IFOUTDISCARDS,BER_Counter32,&ifOutDiscards,"ifOutDiscards");
		IFADD (IFENTRY_IFOUTERRORS,BER_Counter32,&ifOutErrors,"ifOutErrors");

		if (ioctl (fd,SIOCGIFTXQLEN,&ifr) < 0)
		  {
			 log_printf (LOG_ERROR,"%s: ioctl SIOCGIFTXQLEN failed: %s\n",mod->name,strerror (errno));
			 goto fail;
		  }
		ifOutQLen = ifr.ifr_qlen;

		IFADD (IFENTRY_IFOUTQLEN,BER_Gauge32,&ifOutQLen,"ifOutQLen");

		/* ccitt.zeroDotZero */
		ifSpecific[0] = 1;
		ifSpecific[1] = 0;

		IFADD (IFENTRY_IFSPECIFIC,BER_OID,ifSpecific,"ifSpecific");
	 }

   mem_free (stats);
   close (fd);

   return (0);
}

static int system_open (const struct mod_r *mod,const char *confdir)
{
   /* iso.org.dod.internet.mgmt.mib-2.system.xx.xx */
   uint32_t oid[9] = { 8, 43, 6, 1, 2, 1, 1, 0, 0 };
   struct utsname buf;
   octet_string_t str;
   snmp_value_t value;
   /* iso.org.dod.internet.private.enterprises.frogfoot.servers */
   uint32_t sysObjectID[8] = { 7, 43, 6, 1, 4, 1, 10002, 1 };
   int32_t sysServices;
   uint32_t sysORLastChange;
   const mib2_t *mib2;

   if ((mib2 = mib2_load (mod->name,confdir)) == NULL)
	 return (-1);

   if (uname (&buf) < 0)
	 {
		log_printf (LOG_ERROR,"%s: uname failed: %s\n",mod->name,strerror (errno));
		return (-1);
	 }

   str.len = strlen (buf.sysname) + strlen (buf.release) + strlen (buf.version) + strlen (buf.machine) + 3;

   if ((str.buf = (uint8_t *) mem_alloc ((str.len + 1) * sizeof (char))) == NULL)
	 {
		log_printf (LOG_ERROR,"%s: couldn't allocate memory for temporary storage: %s\n",mod->name,strerror (errno));
		return (-1);
	 }

   strcpy (str.buf,buf.sysname);
   strcat (str.buf," ");
   strcat (str.buf,buf.release);
   strcat (str.buf," ");
   strcat (str.buf,buf.version);
   strcat (str.buf," ");
   strcat (str.buf,buf.machine);

   oid[7] = 1;

   if (set_value (&value,BER_OCTET_STRING,&str) < 0 || odb_add (oid,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: couldn't add system.sysDescr.0: %s\n",mod->name,strerror (errno));
		mem_free (str.buf);
		return (-1);
	 }

   mem_free (str.buf);

   oid[7] = 2;

   if (set_value (&value,BER_OID,sysObjectID) < 0 || odb_add (oid,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: couldn't add system.sysObjectID.0: %s\n",mod->name,strerror (errno));
		odb_remove (mod->oid);
		return (-1);
	 }

   str.len = strlen (mib2->contact);
   str.buf = (uint8_t *) mib2->contact;

   oid[7] = 4;

   if (set_value (&value,BER_OCTET_STRING,&str) < 0 || odb_add (oid,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: couldn't add system.sysContact.0: %s\n",mod->name,strerror (errno));
		odb_remove (mod->oid);
		return (-1);
	 }

   str.len = strlen (mib2->location);
   str.buf = (uint8_t *) mib2->location;

   oid[7] = 6;

   if (set_value (&value,BER_OCTET_STRING,&str) < 0 || odb_add (oid,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: couldn't add system.sysLocation.0: %s\n",mod->name,strerror (errno));
		odb_remove (mod->oid);
		return (-1);
	 }

   oid[7] = 7;

#define LAYER(x) (1 << (x))

   /* Linux, FreeBSD, etc. can do layer 1-4, 7 */
   sysServices = LAYER(1) | LAYER(2) | LAYER(3) | LAYER(4) | LAYER(7);

   if (set_value (&value,BER_INTEGER,&sysServices) < 0 || odb_add (oid,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: couldn't add system.sysServices.0: %s\n",mod->name,strerror (errno));
		odb_remove (mod->oid);
		return (-1);
	 }

   oid[7] = 8;

   /* for now, we pretend life started in the beginning */
   sysORLastChange = 0;

   if (set_value (&value,BER_TimeTicks,&sysORLastChange) < 0 || odb_add (oid,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: couldn't add system.sysORLastChange.0: %s\n",mod->name,strerror (errno));
		odb_remove (mod->oid);
		return (-1);
	 }

   return (0);
}

static int system_update (const struct mod_r *mod)
{
   /* iso.org.dod.internet.mgmt.mib-2.system.xx.xx */
   uint32_t oid[9] = { 8, 43, 6, 1, 2, 1, 1, 0, 0 };
   uint8_t hostname[MAXHOSTNAMELEN];
   snmp_value_t value;
   octet_string_t str;
   uint32_t uptime;

   oid[7] = 3;
   odb_remove (oid);

   if (getprocuptime (&uptime) < 0)
	 {
		log_printf (LOG_ERROR,"%s: failed to get process uptime: %s\n",mod->name,strerror (errno));
		return (-1);
	 }

   uptime *= 100;

   if (set_value (&value,BER_TimeTicks,&uptime) < 0 || odb_add (oid,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: couldn't add system.sysUpTime.0: %s\n",mod->name,strerror (errno));
		return (-1);
	 }

   if (gethostname ((char *) hostname,MAXHOSTNAMELEN) < 0)
	 {
		log_printf (LOG_ERROR,"%s: gethostname failed: %s\n",mod->name,strerror (errno));
		return (-1);
	 }
   hostname[MAXHOSTNAMELEN - 1] = '\0';

   str.len = strlen (hostname);
   str.buf = hostname;

   oid[7] = 5;
   odb_remove (oid);
   if (set_value (&value,BER_OCTET_STRING,&str) < 0 || odb_add (oid,&value) < 0)
	 {
		log_printf (LOG_ERROR,"%s: couldn't add system.sysName.0: %s\n",mod->name,strerror (errno));
		return (-1);
	 }

   /*
	* TODO:
	*
	* sysORLastChange (remove dummy from system_open())
	* sysORTable
	*/

   return(0);
}

static void system_close (const struct mod_r *mod)
{
   mib2_close ();
}

/* iso.org.dod.internet.mgmt.mib-2.system */
static const uint32_t system_oid[7] = { 6, 43, 6, 1, 2, 1, 1 };

/* iso.org.dod.internet.mgmt.mib-2.interfaces */
static const uint32_t iface_oid[7] = { 6, 43, 6, 1, 2, 1, 2 };

const mod_t mod[] =
{
   { "mib-2/system", system_open, system_update, system_close, system_oid },
   { "mib-2/interfaces", NULL, iface_update, NULL, iface_oid },
   { NULL, NULL, NULL, NULL, NULL }
};

